import logging
import re
import string
import warnings
from copy import copy
from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING, Generic, Optional, Union
from uuid import uuid4

from slixmpp import JID, Iq, Message, Presence
from slixmpp.exceptions import IqError, XMPPError
from slixmpp.jid import _unescape_node
from slixmpp.plugins.xep_0004 import Form
from slixmpp.plugins.xep_0060.stanza import Item
from slixmpp.plugins.xep_0082 import parse as str_to_datetime
from slixmpp.xmlstream import ET

from ..contact.roster import ContactIsUser
from ..core import config
from ..core.mixins.avatar import AvatarMixin
from ..core.mixins.disco import ChatterDiscoMixin
from ..core.mixins.lock import NamedLockMixin
from ..core.mixins.recipient import ReactionRecipientMixin, ThreadRecipientMixin
from ..util import ABCSubclassableOnceAtMost
from ..util.types import (
    LegacyGroupIdType,
    LegacyMessageType,
    LegacyParticipantType,
    LegacyUserIdType,
    Mention,
    MucAffiliation,
    MucType,
)
from ..util.util import deprecated
from .archive import MessageArchive

if TYPE_CHECKING:
    from ..contact import LegacyContact
    from ..core.gateway import BaseGateway
    from ..core.session import BaseSession

ADMIN_NS = "http://jabber.org/protocol/muc#admin"


class LegacyMUC(
    Generic[
        LegacyGroupIdType, LegacyMessageType, LegacyParticipantType, LegacyUserIdType
    ],
    AvatarMixin,
    NamedLockMixin,
    ChatterDiscoMixin,
    ReactionRecipientMixin,
    ThreadRecipientMixin,
    metaclass=ABCSubclassableOnceAtMost,
):
    """
    A room, a.k.a. a Multi-User Chat.

    MUC instances are obtained by calling :py:meth:`slidge.group.bookmarks.LegacyBookmarks`
    on the user's :py:class:`slidge.core.session.BaseSession`.
    """

    subject_date: Optional[datetime] = None
    n_participants: Optional[int] = None
    max_history_fetch = 100

    type = MucType.CHANNEL
    is_group = True

    DISCO_TYPE = "text"
    DISCO_CATEGORY = "conference"
    DISCO_NAME = "unnamed-room"

    STABLE_ARCHIVE = False
    """
    Because legacy events like reactions, editions, etc. don't all map to a stanza
    with a proper legacy ID, slidge usually cannot guarantee the stability of the archive
    across restarts.

    Set this to True if you know what you're doing, but realistically, this can't
    be set to True until archive is permanently stored on disk by slidge.

    This is just a flag on archive responses that most clients ignore anyway.
    """

    KEEP_BACKFILLED_PARTICIPANTS = False
    """
    Set this to ``True`` if the participant list is not full after calling
    ``fill_participants()``. This is a workaround for networks with huge
    participant lists which do not map really well the MUCs where all presences
    are sent on join.
    It allows to ensure that the participants that last spoke (within the
    ``fill_history()`` method are effectively participants, thus making possible
    for XMPP clients to fetch their avatars.
    """

    _ALL_INFO_FILLED_ON_STARTUP = False
    """
    Set this to true if the fill_participants() / fill_participants() design does not
    fit the legacy API, ie, no lazy loading of the participant list and history.
    """

    HAS_DESCRIPTION = True
    """
    Set this to false if the legacy network does not allow setting a description
    for the group. In this case the description field will not be present in the
    room configuration form.
    """

    HAS_SUBJECT = True
    """
    Set this to false if the legacy network does not allow setting a subject
    (sometimes also called topic) for the group. In this case, as a subject is
    recommended by :xep:`0045` ("SHALL"), the description (or the group name as
    ultimate fallback) will be used as the room subject.
    By setting this to false, an error will be returned when the :term:`User`
    tries to set the room subject.
    """

    _avatar_pubsub_broadcast = False
    _avatar_bare_jid = True

    def __init__(self, session: "BaseSession", legacy_id: LegacyGroupIdType, jid: JID):
        from .participant import LegacyParticipant

        self.session = session
        self.xmpp: "BaseGateway" = session.xmpp
        self.user = session.user
        self.log = logging.getLogger(f"{self.user.bare_jid}:muc:{jid}")

        self.legacy_id = legacy_id
        self.jid = jid

        self.user_resources = set[str]()

        self.Participant = LegacyParticipant.get_self_or_unique_subclass()

        self.xmpp.add_event_handler(
            "presence_unavailable", self._on_presence_unavailable
        )

        self._subject = ""
        self.subject_setter: Union[str, "LegacyContact", "LegacyParticipant"] = (
            "unknown"
        )

        self.archive: MessageArchive = MessageArchive(str(self.jid), self.user)
        self._user_nick: Optional[str] = None

        self._participants_by_nicknames = dict[str, LegacyParticipantType]()
        self._participants_by_escaped_nicknames = dict[str, LegacyParticipantType]()
        self._participants_by_contacts = dict["LegacyContact", LegacyParticipantType]()

        self.__participants_filled = False
        self.__history_filled = False
        self._description = ""
        super().__init__()

    def __repr__(self):
        return f"<MUC {self.legacy_id}/{self.jid}/{self.name}>"

    def __send_configuration_change(self, codes):
        part = self.get_system_participant()
        part.send_configuration_change(codes)

    @property
    def user_nick(self):
        return (
            self._user_nick
            or self.session.bookmarks.user_nick
            or self.session.user.jid.node
        )

    @user_nick.setter
    def user_nick(self, nick: str):
        self._user_nick = nick

    async def __fill_participants(self):
        async with self.lock("fill participants"):
            if self.__participants_filled:
                return
            self.__participants_filled = True
            try:
                await self.fill_participants()
            except NotImplementedError:
                pass

    async def __fill_history(self):
        async with self.lock("fill history"):
            if self.__history_filled:
                log.debug("History has already been fetched %s", self)
                return
            log.debug("Fetching history for %s", self)
            for msg in self.archive:
                try:
                    legacy_id = self.session.xmpp_to_legacy_msg_id(msg.id)
                    oldest_date = msg.when
                except Exception as e:
                    # not all archived stanzas have a valid legacy msg ID, eg
                    # reactions, corrections, message with multiple attachments…
                    self.log.debug(f"Could not convert during history back-filling {e}")
                else:
                    break
            else:
                legacy_id = None
                oldest_date = None
            try:
                await self.backfill(legacy_id, oldest_date)
            except NotImplementedError:
                return
            self.__history_filled = True

    @property
    def name(self):
        return self.DISCO_NAME

    @name.setter
    def name(self, n: str):
        if self.DISCO_NAME == n:
            return
        self.DISCO_NAME = n
        self.__send_configuration_change((104,))

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, d: str):
        if self._description == d:
            return
        self._description = d
        self.__send_configuration_change((104,))

    def _on_presence_unavailable(self, p: Presence):
        pto = p.get_to()
        if pto.bare != self.jid.bare:
            return

        pfrom = p.get_from()
        if pfrom.bare != self.user.bare_jid:
            return
        if (resource := pfrom.resource) in (resources := self.user_resources):
            if pto.resource != self.user_nick:
                self.log.debug(
                    "Received 'leave group' request but with wrong nickname. %s", p
                )
            resources.remove(resource)
        else:
            self.log.debug(
                "Received 'leave group' request but resource was not listed. %s", p
            )

    async def update_info(self):
        """
        Fetch information about this group from the legacy network

        This is awaited on MUC instantiation, and should be overridden to
        update the attributes of the group chat, like title, subject, number
        of participants etc.

        To take advantage of the slidge avatar cache, you can check the .avatar
        property to retrieve the "legacy file ID" of the cached avatar. If there
        is no change, you should not call
        :py:meth:`slidge.core.mixins.avatar.AvatarMixin.set_avatar()` or
        attempt to modify
        the :attr:.avatar property.
        """
        raise NotImplementedError

    async def backfill(
        self,
        oldest_message_id: Optional[LegacyMessageType] = None,
        oldest_message_date: Optional[datetime] = None,
    ):
        """
        Override this if the legacy network provide server-side archive.
        In it, send history messages using ``self.get_participant().send*``,
        with the ``archive_only=True`` kwarg.

        You only need to fetch messages older than ``oldest_message_id``.

        :param oldest_message_id: The oldest message ID already present in the archive
        :param oldest_message_date: The oldest message date already present in the archive
        """
        raise NotImplementedError

    async def fill_participants(self):
        """
        In here, call self.get_participant(), self.get_participant_by_contact(),
        of self.get_user_participant() to make an initial list of participants.
        """
        raise NotImplementedError

    @property
    def subject(self):
        return self._subject

    @subject.setter
    def subject(self, s: str):
        if s == self._subject:
            return
        self.xmpp.loop.create_task(
            self.__get_subject_setter_participant()
        ).add_done_callback(
            lambda task: task.result().set_room_subject(
                s, None, self.subject_date, False
            )
        )
        self._subject = s

    @property
    def is_anonymous(self):
        return self.type == MucType.CHANNEL

    async def __get_subject_setter_participant(self):
        from slidge.contact import LegacyContact

        from .participant import LegacyParticipant

        who = self.subject_setter

        if isinstance(who, LegacyParticipant):
            return who
        elif isinstance(who, str):
            return await self.get_participant(who, store=False)
        elif isinstance(self.subject_setter, LegacyContact):
            return await self.get_participant_by_contact(who)
        else:
            return self.get_system_participant()

    def features(self):
        features = [
            "http://jabber.org/protocol/muc",
            "http://jabber.org/protocol/muc#stable_id",
            "http://jabber.org/protocol/muc#self-ping-optimization",
            "urn:xmpp:mam:2",
            "urn:xmpp:mam:2#extended",
            "urn:xmpp:sid:0",
            "muc_persistent",
            "vcard-temp",
            "urn:xmpp:ping",
            "urn:xmpp:occupant-id:0",
            self.xmpp.plugin["xep_0425"].stanza.NS,
        ]
        if self.type == MucType.GROUP:
            features.extend(["muc_membersonly", "muc_nonanonymous", "muc_hidden"])
        elif self.type == MucType.CHANNEL:
            features.extend(["muc_open", "muc_semianonymous", "muc_public"])
        elif self.type == MucType.CHANNEL_NON_ANONYMOUS:
            features.extend(["muc_open", "muc_nonanonymous", "muc_public"])
        return features

    async def extended_features(self):
        is_group = self.type == MucType.GROUP

        form = self.xmpp.plugin["xep_0004"].make_form(ftype="result")

        form.add_field(
            "FORM_TYPE", "hidden", value="http://jabber.org/protocol/muc#roominfo"
        )
        form.add_field("muc#roomconfig_persistentroom", "boolean", value=True)
        form.add_field("muc#roomconfig_changesubject", "boolean", value=False)
        form.add_field("muc#maxhistoryfetch", value=str(self.max_history_fetch))
        form.add_field("muc#roominfo_subjectmod", "boolean", value=False)

        if self._ALL_INFO_FILLED_ON_STARTUP or self.__participants_filled:
            n: Optional[int] = len(await self.get_participants())
        else:
            n = self.n_participants
        if n is not None:
            form.add_field("muc#roominfo_occupants", value=str(n))

        if d := self.description:
            form.add_field("muc#roominfo_description", value=d)

        if s := self.subject:
            form.add_field("muc#roominfo_subject", value=s)

        if self._set_avatar_task:
            await self._set_avatar_task
            avatar = self.get_avatar()
            if avatar and (h := avatar.id):
                form.add_field(
                    "{http://modules.prosody.im/mod_vcard_muc}avatar#sha1", value=h
                )
                form.add_field("muc#roominfo_avatarhash", "text-multi", value=[h])

        form.add_field("muc#roomconfig_membersonly", "boolean", value=is_group)
        form.add_field(
            "muc#roomconfig_whois",
            "list-single",
            value="moderators" if self.is_anonymous else "anyone",
        )
        form.add_field("muc#roomconfig_publicroom", "boolean", value=not is_group)
        form.add_field("muc#roomconfig_allowpm", "boolean", value=False)

        r = [form]

        if reaction_form := await self.restricted_emoji_extended_feature():
            r.append(reaction_form)

        return r

    def shutdown(self):
        user_jid = copy(self.jid)
        user_jid.resource = self.user_nick
        for user_full_jid in self.user_full_jids():
            presence = self.xmpp.make_presence(
                pfrom=user_jid, pto=user_full_jid, ptype="unavailable"
            )
            presence["muc"]["affiliation"] = "none"
            presence["muc"]["role"] = "none"
            presence["muc"]["status_codes"] = {110, 332}
            presence.send()

    def user_full_jids(self):
        for r in self.user_resources:
            j = copy(self.user.jid)
            j.resource = r
            yield j

    @property
    def user_muc_jid(self):
        user_muc_jid = copy(self.jid)
        user_muc_jid.resource = self.user_nick
        return user_muc_jid

    def _legacy_to_xmpp(self, legacy_id: LegacyMessageType):
        return self.session.sent.get(legacy_id) or self.session.legacy_to_xmpp_msg_id(
            legacy_id
        )

    async def echo(
        self, msg: Message, legacy_msg_id: Optional[LegacyMessageType] = None
    ):
        origin_id = msg.get_origin_id()

        msg.set_from(self.user_muc_jid)
        msg.set_id(msg.get_id())
        if origin_id:
            # because of slixmpp internal magic, we need to do this to ensure the origin_id
            # is present
            set_origin_id(msg, origin_id)
        if legacy_msg_id:
            msg["stanza_id"]["id"] = self.session.legacy_to_xmpp_msg_id(legacy_msg_id)
        else:
            msg["stanza_id"]["id"] = str(uuid4())
        msg["stanza_id"]["by"] = self.jid
        msg["occupant-id"]["id"] = "slidge-user"

        self.archive.add(msg, await self.get_user_participant())

        for user_full_jid in self.user_full_jids():
            self.log.debug("Echoing to %s", user_full_jid)
            msg = copy(msg)
            msg.set_to(user_full_jid)

            msg.send()

    def _post_avatar_update(self) -> None:
        self.__send_configuration_change((104,))
        self._send_room_presence()

    def _send_room_presence(self, user_full_jid: Optional[JID] = None):
        if user_full_jid is None:
            tos = self.user_full_jids()
        else:
            tos = [user_full_jid]
        for to in tos:
            p = self.xmpp.make_presence(pfrom=self.jid, pto=to)
            if (avatar := self.get_avatar()) and (h := avatar.id):
                p["vcard_temp_update"]["photo"] = h
            else:
                p["vcard_temp_update"]["photo"] = ""
            p.send()

    async def join(self, join_presence: Presence):
        user_full_jid = join_presence.get_from()
        requested_nickname = join_presence.get_to().resource
        client_resource = user_full_jid.resource

        if client_resource in self.user_resources:
            self.log.debug("Received join from a resource that is already joined.")

        self.user_resources.add(client_resource)

        if not requested_nickname or not client_resource:
            raise XMPPError("jid-malformed", by=self.jid)

        self.log.debug(
            "Resource %s of %s wants to join room %s with nickname %s",
            client_resource,
            self.user,
            self.legacy_id,
            requested_nickname,
        )

        await self.__fill_participants()

        for participant in self._participants_by_nicknames.values():
            if participant.is_user:  # type:ignore
                continue
            if participant.is_system:  # type:ignore
                continue
            participant.send_initial_presence(full_jid=user_full_jid)

        user_nick = self.user_nick
        user_participant = await self.get_user_participant()
        if not user_participant.is_user:  # type:ignore
            self.log.warning("is_user flag not set participant on user_participant")
            user_participant.is_user = True  # type:ignore
        user_participant.send_initial_presence(
            user_full_jid,
            presence_id=join_presence["id"],
            nick_change=user_nick != requested_nickname,
        )

        history_params = join_presence["muc_join"]["history"]
        maxchars = int_or_none(history_params["maxchars"])
        maxstanzas = int_or_none(history_params["maxstanzas"])
        seconds = int_or_none(history_params["seconds"])
        try:
            since = self.xmpp.plugin["xep_0082"].parse(history_params["since"])
        except ValueError:
            since = None
        if seconds:
            since = datetime.now() - timedelta(seconds=seconds)
        if equals_zero(maxchars) or equals_zero(maxstanzas):
            log.debug("Joining client does not want any old-school MUC history-on-join")
        else:
            self.log.debug("Old school history fill")
            await self.__fill_history()
            await self.__old_school_history(
                user_full_jid,
                maxchars=maxchars,
                maxstanzas=maxstanzas,
                since=since,
            )
        (await self.__get_subject_setter_participant()).set_room_subject(
            self._subject if self.HAS_SUBJECT else (self.description or self.name),
            user_full_jid,
            self.subject_date,
        )
        if t := self._set_avatar_task:
            await t
        self._send_room_presence(user_full_jid)
        self.user_resources.add(client_resource)

    async def get_user_participant(self, **kwargs) -> "LegacyParticipantType":
        """
        Get the participant representing the gateway user

        :param kwargs: additional parameters for the :class:`.Participant`
            construction (optional)
        :return:
        """
        p = await self.get_participant(self.user_nick, is_user=True, **kwargs)
        self.__store_participant(p)
        return p

    def __store_participant(self, p: "LegacyParticipantType"):
        # we don't want to update the participant list when we're filling history
        if not self.KEEP_BACKFILLED_PARTICIPANTS and self.get_lock("fill history"):
            return
        self._participants_by_nicknames[p.nickname] = p  # type:ignore
        if p.contact:
            self._participants_by_contacts[p.contact] = p

    async def get_participant(
        self,
        nickname: str,
        raise_if_not_found=False,
        fill_first=False,
        store=True,
        **kwargs,
    ) -> "LegacyParticipantType":
        """
        Get a participant by their nickname.

        In non-anonymous groups, you probably want to use
        :meth:`.LegacyMUC.get_participant_by_contact` instead.

        :param nickname: Nickname of the participant (used as resource part in the MUC)
        :param raise_if_not_found: Raise XMPPError("item-not-found") if they are not
            in the participant list (internal use by slidge, plugins should not
            need that)
        :param fill_first: Ensure :meth:`.LegacyMUC.fill_participants()` has been called first
             (internal use by slidge, plugins should not need that)
        :param store: persistently store the user in the list of MUC participants
        :param kwargs: additional parameters for the :class:`.Participant`
            construction (optional)
        :return:
        """
        if fill_first:
            await self.__fill_participants()
        p = self._participants_by_nicknames.get(
            nickname
        ) or self._participants_by_escaped_nicknames.get(nickname)
        if p is None:
            if raise_if_not_found:
                raise XMPPError("item-not-found")
            p = self.Participant(self, nickname, **kwargs)
            if store:
                self.__store_participant(p)
        return p

    def get_system_participant(self) -> "LegacyParticipantType":
        """
        Get a pseudo-participant, representing the room itself

        Can be useful for events that cannot be mapped to a participant,
        e.g. anonymous moderation events, or announces from the legacy
        service
        :return:
        """
        return self.Participant(self, is_system=True)

    async def get_participant_by_contact(
        self, c: "LegacyContact", **kwargs
    ) -> "LegacyParticipantType":
        """
        Get a non-anonymous participant.

        This is what should be used in non-anonymous groups ideally, to ensure
        that the Contact jid is associated to this participant

        :param c: The :class:`.LegacyContact` instance corresponding to this contact
        :param kwargs: additional parameters for the :class:`.Participant`
            construction (optional)
        :return:
        """
        await self.session.contacts.ready
        p = self._participants_by_contacts.get(c)
        if p is None:
            nickname = c.name or _unescape_node(c.jid_username)
            if nickname in self._participants_by_nicknames:
                self.log.debug("Nickname conflict")
                nickname = f"{nickname} ({c.jid_username})"
            p = self.Participant(self, nickname, **kwargs)
            p.contact = c
            c.participants.add(p)
            # FIXME: this is not great but given the current design,
            #        during participants fill and history backfill we do not
            #        want to send presence, because we might update affiliation
            #        and role afterwards.
            # We need a refactor of the MUC class… later™
            if not self.get_lock("fill participants") and not self.get_lock(
                "fill history"
            ):
                p.send_last_presence(force=True, no_cache_online=True)
            self.__store_participant(p)
        return p

    async def get_participant_by_legacy_id(
        self, legacy_id: LegacyUserIdType, **kwargs
    ) -> "LegacyParticipantType":
        try:
            c = await self.session.contacts.by_legacy_id(legacy_id)
        except ContactIsUser:
            return await self.get_user_participant(**kwargs)
        return await self.get_participant_by_contact(c, **kwargs)

    async def get_participants(self):
        """
        Get all known participants of the group, ensure :meth:`.LegacyMUC.fill_participants`
        has been awaited once before. Plugins should not use that, internal
        slidge use only.
        :return:
        """
        await self.__fill_participants()
        return list(self._participants_by_nicknames.values())

    def remove_participant(self, p: "LegacyParticipantType", kick=False, ban=False):
        """
        Call this when a participant leaves the room

        :param p: The participant
        :param kick: Whether the participant left because they were kicked
        :param ban: Whether the participant left because they were banned
        """
        if kick and ban:
            raise TypeError("Either kick or ban")
        if p.contact is not None:
            try:
                del self._participants_by_contacts[p.contact]
            except KeyError:
                self.log.warning(
                    "Removed a participant we didn't know was here?, %s", p
                )
            else:
                p.contact.participants.remove(p)
        try:
            del self._participants_by_nicknames[p.nickname]  # type:ignore
        except KeyError:
            self.log.warning("Removed a participant we didn't know was here?, %s", p)
        if kick:
            codes = {307}
        elif ban:
            codes = {301}
        else:
            codes = None
        presence = p._make_presence(ptype="unavailable", status_codes=codes)
        p._affiliation = "outcast" if ban else "none"
        p._role = "none"
        p._send(presence)

    def rename_participant(self, old_nickname: str, new_nickname: str):
        try:
            p = self._participants_by_nicknames.pop(old_nickname)
        except KeyError:
            # when called by participant.nickname.setter
            return
        self._participants_by_nicknames[new_nickname] = p
        if p.nickname == old_nickname:
            p.nickname = new_nickname

    async def __old_school_history(
        self,
        full_jid: JID,
        maxchars: Optional[int] = None,
        maxstanzas: Optional[int] = None,
        seconds: Optional[int] = None,
        since: Optional[datetime] = None,
    ):
        """
        Old-style history join (internal slidge use)

        :param full_jid:
        :param maxchars:
        :param maxstanzas:
        :param seconds:
        :param since:
        :return:
        """
        if since is None:
            if seconds is None:
                start_date = datetime.now(tz=timezone.utc) - timedelta(days=1)
            else:
                start_date = datetime.now(tz=timezone.utc) - timedelta(seconds=seconds)
        else:
            start_date = since or datetime.now(tz=timezone.utc) - timedelta(days=1)

        for h_msg in self.archive.get_all(
            start_date=start_date, end_date=None, last_page_n=maxstanzas
        ):
            msg = h_msg.stanza_component_ns
            msg["delay"]["stamp"] = h_msg.when
            msg.set_to(full_jid)
            self.xmpp.send(msg, False)

    async def send_mam(self, iq: Iq):
        await self.__fill_history()

        form_values = iq["mam"]["form"].get_values()

        start_date = str_to_datetime_or_none(form_values.get("start"))
        end_date = str_to_datetime_or_none(form_values.get("end"))

        after_id = form_values.get("after-id")
        before_id = form_values.get("before-id")

        sender = form_values.get("with")

        ids = form_values.get("ids") or ()

        if max_str := iq["mam"]["rsm"]["max"]:
            try:
                max_results = int(max_str)
            except ValueError:
                max_results = None
        else:
            max_results = None

        after_id_rsm = iq["mam"]["rsm"]["after"]
        after_id = after_id_rsm or after_id

        before_rsm = iq["mam"]["rsm"]["before"]
        if before_rsm is True and max_results is not None:
            last_page_n = max_results
        else:
            last_page_n = None

        first = None
        last = None
        count = 0

        it = self.archive.get_all(
            start_date,
            end_date,
            before_id,
            after_id,
            ids,
            last_page_n,
            sender,
            bool(iq["mam"]["flip_page"]),
        )

        for history_msg in it:
            last = xmpp_id = history_msg.id
            if first is None:
                first = xmpp_id

            wrapper_msg = self.xmpp.make_message(mfrom=self.jid, mto=iq.get_from())
            wrapper_msg["mam_result"]["queryid"] = iq["mam"]["queryid"]
            wrapper_msg["mam_result"]["id"] = xmpp_id
            wrapper_msg["mam_result"].append(history_msg.forwarded())

            wrapper_msg.send()
            count += 1

            if max_results and count == max_results:
                break

        if max_results:
            try:
                next(it)
            except StopIteration:
                complete = True
            else:
                complete = False
        else:
            complete = True

        reply = iq.reply()
        if not self.STABLE_ARCHIVE:
            reply["mam_fin"]["stable"] = "false"
        if complete:
            reply["mam_fin"]["complete"] = "true"
        reply["mam_fin"]["rsm"]["first"] = first
        reply["mam_fin"]["rsm"]["last"] = last
        reply["mam_fin"]["rsm"]["count"] = str(count)
        reply.send()

    async def send_mam_metadata(self, iq: Iq):
        await self.__fill_history()
        await self.archive.send_metadata(iq)

    async def kick_resource(self, r: str):
        """
        Kick a XMPP client of the user. (slidge internal use)

        :param r: The resource to kick
        """
        pto = self.user.jid
        pto.resource = r
        p = self.xmpp.make_presence(
            pfrom=(await self.get_user_participant()).jid, pto=pto
        )
        p["muc"]["affiliation"] = "none"
        p["muc"]["role"] = "none"
        p["muc"]["status_codes"] = {110, 333}
        p.send()

    async def add_to_bookmarks(self, auto_join=True, invite=False, preserve=True):
        """
        Add the MUC to the user's XMPP bookmarks (:xep:`0402')

        This requires that slidge has the IQ privileged set correctly
        on the XMPP server

        :param auto_join: whether XMPP clients should automatically join
            this MUC on startup. In theory, XMPP clients will receive
            a "push" notification when this is called, and they will
            join if they are online.
        :param invite: send an invitation to join this MUC emanating from
            the gateway. While this should not be strictly necessary,
            it can help for clients that do not support :xep:`0402`, or
            that have 'do not honor bookmarks auto-join' turned on in their
            settings.
        :param preserve: preserve auto-join and bookmarks extensions
            set by the user outside slidge
        """
        item = Item()
        item["id"] = self.jid

        iq = Iq(stype="get", sfrom=self.user.jid, sto=self.user.jid)
        iq["pubsub"]["items"]["node"] = self.xmpp["xep_0402"].stanza.NS
        iq["pubsub"]["items"].append(item)

        is_update = False
        if preserve:
            try:
                ans = await self.xmpp["xep_0356"].send_privileged_iq(iq)
                is_update = len(ans["pubsub"]["items"]) == 1
                # this below creates the item if it wasn't here already
                # (slixmpp annoying magic)
                item = ans["pubsub"]["items"]["item"]
                item["id"] = self.jid
            except IqError:
                item["conference"]["autojoin"] = auto_join
            except PermissionError:
                warnings.warn(
                    "IQ privileges (XEP0356) are not set, we cannot fetch the user bookmarks"
                )
            else:
                # if the bookmark is already present, we preserve it as much as
                # possible, especially custom <extensions>
                self.log.debug("Existing: %s", item)
                # if it's an update, we do not touch the auto join flag
                if not is_update:
                    item["conference"]["autojoin"] = auto_join
        else:
            item["conference"]["autojoin"] = auto_join

        item["conference"]["nick"] = self.user_nick
        iq = Iq(stype="set", sfrom=self.user.jid, sto=self.user.jid)
        iq["pubsub"]["publish"]["node"] = self.xmpp["xep_0402"].stanza.NS
        iq["pubsub"]["publish"].append(item)

        iq["pubsub"]["publish_options"] = _BOOKMARKS_OPTIONS

        try:
            await self.xmpp["xep_0356"].send_privileged_iq(iq)
        except PermissionError:
            warnings.warn(
                "IQ privileges (XEP0356) are not set, we cannot add bookmarks for the user"
            )
            # fallback by forcing invitation
            invite = True
        except IqError as e:
            warnings.warn(
                f"Something went wrong while trying to set the bookmarks: {e}"
            )
            # fallback by forcing invitation
            invite = True

        if invite or (config.ALWAYS_INVITE_WHEN_ADDING_BOOKMARKS and not is_update):
            self.session.send_gateway_invite(
                self, reason="This group could not be added automatically for you"
            )

    async def on_avatar(
        self, data: Optional[bytes], mime: Optional[str]
    ) -> Optional[Union[int, str]]:
        """
        Called when the user tries to set the avatar of the room from an XMPP
        client.

        If the set avatar operation is completed, should return a legacy image
        unique identifier. In this case the MUC avatar will be immediately
        updated on the XMPP side.

        If data is not None and this method returns None, then we assume that
        self.set_avatar() will be called elsewhere, eg triggered by a legacy
        room update event.

        :param data: image data or None if the user meant to remove the avatar
        :param mime: the mime type of the image. Since this is provided by
            the XMPP client, there is no guarantee that this is valid or
            correct.
        :return: A unique avatar identifier, which will trigger
            :py:meth:`slidge.group.room.LegacyMUC.set_avatar`. Alternatively, None, if
            :py:meth:`.LegacyMUC.set_avatar` is meant to be awaited somewhere else.
        """
        raise NotImplementedError

    admin_set_avatar = deprecated("LegacyMUC.on_avatar", on_avatar)

    async def on_set_affiliation(
        self,
        contact: "LegacyContact",
        affiliation: MucAffiliation,
        reason: Optional[str],
        nickname: Optional[str],
    ):
        """
        Triggered when the user requests changing the affiliation of a contact
        for this group,

        Examples: promotion them to moderator, kick (affiliation=none),
        ban (affiliation=outcast).

        :param contact: The contact whose affiliation change is requested
        :param affiliation: The new affiliation
        :param reason: A reason for this affiliation change
        :param nickname:
        """
        raise NotImplementedError

    async def on_set_config(
        self,
        name: Optional[str],
        description: Optional[str],
    ):
        """
        Triggered when the user requests changing the room configuration.
        Only title and description can be changed at the moment.

        The legacy module is responsible for updating :attr:`.title` and/or
        :attr:`.description` of this instance.

        If :attr:`.HAS_DESCRIPTION` is set to False, description will always
        be ``None``.

        :param name: The new name of the room.
        :param description: The new description of the room.
        """
        raise NotImplementedError

    async def on_destroy_request(self, reason: Optional[str]):
        """
        Triggered when the user requests room destruction.

        :param reason: Optionally, a reason for the destruction
        """
        raise NotImplementedError

    async def parse_mentions(self, text: str) -> list[Mention]:
        await self.__fill_participants()

        if len(self._participants_by_nicknames) == 0:
            return []

        result = []
        for match in re.finditer(
            "|".join(
                sorted(
                    [re.escape(nick) for nick in self._participants_by_nicknames],
                    key=lambda nick: len(nick),
                    reverse=True,
                )
            ),
            text,
        ):
            span = match.span()
            nick = match.group()
            if span[0] != 0 and text[span[0] - 1] not in _WHITESPACE_OR_PUNCTUATION:
                continue
            if span[1] == len(text) or text[span[1]] in _WHITESPACE_OR_PUNCTUATION:
                participant = self._participants_by_nicknames[nick]
                if contact := participant.contact:
                    result.append(Mention(contact=contact, start=span[0], end=span[1]))
        return result

    async def on_set_subject(self, subject: str) -> None:
        """
        Triggered when the user requests changing the room subject.

        The legacy module is responsible for updating :attr:`.subject` of this
        instance.

        :param subject: The new subject for this room.
        """
        raise NotImplementedError


def set_origin_id(msg: Message, origin_id: str):
    sub = ET.Element("{urn:xmpp:sid:0}origin-id")
    sub.attrib["id"] = origin_id
    msg.xml.append(sub)


def int_or_none(x):
    try:
        return int(x)
    except ValueError:
        return None


def equals_zero(x):
    if x is None:
        return False
    else:
        return x == 0


def str_to_datetime_or_none(date: Optional[str]):
    if date is None:
        return
    try:
        return str_to_datetime(date)
    except ValueError:
        return None


def bookmarks_form():
    form = Form()
    form["type"] = "submit"
    form.add_field(
        "FORM_TYPE",
        value="http://jabber.org/protocol/pubsub#publish-options",
        ftype="hidden",
    )
    form.add_field("pubsub#persist_items", value="1")
    form.add_field("pubsub#max_items", value="max")
    form.add_field("pubsub#send_last_published_item", value="never")
    form.add_field("pubsub#access_model", value="whitelist")
    return form


_BOOKMARKS_OPTIONS = bookmarks_form()
_WHITESPACE_OR_PUNCTUATION = string.whitespace + string.punctuation

log = logging.getLogger(__name__)



"""
The main slidge package.

Contains importable classes for a minimal function :term:`Legacy Module`.
"""

import sys
import warnings

from . import slixfix  # noqa: F401
from .command import FormField, SearchResult  # noqa: F401
from .contact import LegacyContact, LegacyRoster  # noqa: F401
from .core import config as global_config  # noqa: F401
from .core.gateway import BaseGateway  # noqa: F401
from .core.session import BaseSession  # noqa: F401
from .group import LegacyBookmarks, LegacyMUC, LegacyParticipant  # noqa: F401
from .util.db import GatewayUser, user_store  # noqa: F401
from .util.types import MucType  # noqa: F401
from .util.util import addLoggingLevel

from .__main__ import main  # isort: skip


def entrypoint(module_name: str) -> None:
    """
    Entrypoint to be used in ``__main__.py`` of
    :term:`legacy modules <Legacy Module>`.

    :param module_name: An importable :term:`Legacy Module`.
    """
    sys.argv.extend(["--legacy", module_name])
    main()


def formatwarning(message, category, filename, lineno, line=""):
    return f"{filename}:{lineno}:{category.__name__}:{message}\n"


warnings.formatwarning = formatwarning


__all__ = [
    "BaseGateway",
    "BaseSession",
    # For backwards compatibility, these names are still importable from the
    # top-level slidge module, but this is deprecated.
    # "GatewayUser",
    # "LegacyBookmarks",
    # "LegacyMUC",
    # "LegacyContact",
    # "LegacyParticipant",
    # "LegacyRoster",
    # "MucType",
    # "FormField",
    # "SearchResult",
    "entrypoint",
    "user_store",
    "global_config",
]

addLoggingLevel()



"""
This module extends slixmpp.ComponentXMPP to make writing new LegacyClients easier
"""

import asyncio
import logging
import re
import tempfile
from copy import copy
from datetime import datetime
from typing import TYPE_CHECKING, Callable, Collection, Optional, Sequence, Union

import aiohttp
import qrcode
from slixmpp import JID, ComponentXMPP, Iq, Message, Presence
from slixmpp.exceptions import IqError, IqTimeout, XMPPError
from slixmpp.plugins.xep_0060.stanza import OwnerAffiliation
from slixmpp.types import MessageTypes
from slixmpp.xmlstream.xmlstream import NotConnectedError

from ... import command  # noqa: F401
from ...command.adhoc import AdhocProvider
from ...command.admin import Exec
from ...command.base import Command, FormField
from ...command.chat_command import ChatCommandProvider
from ...command.register import RegistrationType
from ...slixfix.roster import RosterBackend
from ...slixfix.xep_0292.vcard4 import VCard4Provider
from ...util import ABCSubclassableOnceAtMost
from ...util.db import GatewayUser, user_store
from ...util.sql import db
from ...util.types import AvatarType, MessageOrPresenceTypeVar
from .. import config
from ..mixins import MessageMixin
from ..pubsub import PubSubComponent
from ..session import BaseSession
from .caps import Caps
from .delivery_receipt import DeliveryReceipt
from .disco import Disco
from .mam import Mam
from .muc_admin import MucAdmin
from .ping import Ping
from .presence import PresenceHandlerMixin
from .registration import Registration
from .search import Search
from .session_dispatcher import SessionDispatcher
from .vcard_temp import VCardTemp

if TYPE_CHECKING:
    from ...group.room import LegacyMUC


class BaseGateway(
    PresenceHandlerMixin,
    ComponentXMPP,
    MessageMixin,
    metaclass=ABCSubclassableOnceAtMost,
):
    """
    The gateway component, handling registrations and un-registrations.

    On slidge launch, a singleton is instantiated, and it will be made available
    to public classes such :class:`.LegacyContact` or :class:`.BaseSession` as the
    ``.xmpp`` attribute.

    Must be subclassed by a legacy module to set up various aspects of the XMPP
    component behaviour, such as its display name or welcome message, via
    class attributes :attr:`.COMPONENT_NAME` :attr:`.WELCOME_MESSAGE`.

    Abstract methods related to the registration process must be overriden
    for a functional :term:`Legacy Module`:

    - :meth:`.validate`
    - :meth:`.validate_two_factor_code`
    - :meth:`.get_qr_text`
    - :meth:`.confirm_qr`

    NB: Not all of these must be overridden, it depends on the
    :attr:`REGISTRATION_TYPE`.

    The other methods, such as :meth:`.send_text` or :meth:`.react` are the same
    as those of :class:`.LegacyContact` and :class:`.LegacyParticipant`, because
    the component itself is also a "messaging actor", ie, an :term:`XMPP Entity`.
    For these methods, you need to specify the JID of the recipient with the
    `mto` parameter.

    Since it inherits from :class:`slixmpp.componentxmpp.ComponentXMPP`,you also
    have a hand on low-level XMPP interactions via slixmpp methods, e.g.:

    .. code-block:: python

        self.send_presence(
            pfrom="somebody@component.example.com",
            pto="someonwelse@anotherexample.com",
        )

    However, you should not need to do so often since the classes of the plugin
    API provides higher level abstractions around most commonly needed use-cases, such
    as sending messages, or displaying a custom status.

    """

    COMPONENT_NAME: str = NotImplemented
    """Name of the component, as seen in service discovery by XMPP clients"""
    COMPONENT_TYPE: str = ""
    """Type of the gateway, should follow https://xmpp.org/registrar/disco-categories.html"""
    COMPONENT_AVATAR: Optional[AvatarType] = None
    """
    Path, bytes or URL used by the component as an avatar.
    """

    REGISTRATION_FIELDS: Collection[FormField] = [
        FormField(var="username", label="User name", required=True),
        FormField(var="password", label="Password", required=True, private=True),
    ]
    """
    Iterable of fields presented to the gateway user when registering using :xep:`0077`
    `extended <https://xmpp.org/extensions/xep-0077.html#extensibility>`_ by :xep:`0004`.
    """
    REGISTRATION_INSTRUCTIONS: str = "Enter your credentials"
    """
    The text presented to a user that wants to register (or modify) their legacy account
    configuration.
    """
    REGISTRATION_TYPE: RegistrationType = RegistrationType.SINGLE_STEP_FORM
    """
    This attribute determines how users register to the gateway, ie, how they
    login to the :term:`legacy service <Legacy Service>`.
    The credentials are then stored persistently, so this process should happen
    once per user (unless they unregister).

    The registration process always start with a basic data form (:xep:`0004`)
    presented to the user.
    But the legacy login flow might require something more sophisticated, see
    :class:`.RegistrationType` for more details.
    """

    REGISTRATION_2FA_TITLE = "Enter your 2FA code"
    REGISTRATION_2FA_INSTRUCTIONS = (
        "You should have received something via email or SMS, or something"
    )
    REGISTRATION_QR_INSTRUCTIONS = "Flash this code or follow this link"

    ROSTER_GROUP: str = "slidge"
    """
    Name of the group assigned to a :class:`.LegacyContact` automagically
    added to the :term:`User`'s roster with :meth:`.LegacyContact.add_to_roster`.
    """
    WELCOME_MESSAGE = (
        "Thank you for registering. Type 'help' to list the available commands, "
        "or just start messaging away!"
    )
    """
    A welcome message displayed to users on registration.
    This is useful notably for clients that don't consider component JIDs as a
    valid recipient in their UI, yet still open a functional chat window on
    incoming messages from components.
    """

    SEARCH_FIELDS: Sequence[FormField] = [
        FormField(var="first", label="First name", required=True),
        FormField(var="last", label="Last name", required=True),
        FormField(var="phone", label="Phone number", required=False),
    ]
    """
    Fields used for searching items via the component, through :xep:`0055` (jabber search).
    A common use case is to allow users to search for legacy contacts by something else than
    their usernames, eg their phone number.

    Plugins should implement search by overriding :meth:`.BaseSession.search`
    (restricted to registered users).

    If there is only one field, it can also be used via the ``jabber:iq:gateway`` protocol
    described in :xep:`0100`. Limitation: this only works if the search request returns
    one result item, and if this item has a 'jid' var.
    """
    SEARCH_TITLE: str = "Search for legacy contacts"
    """
    Title of the search form.
    """
    SEARCH_INSTRUCTIONS: str = ""
    """
    Instructions of the search form.
    """

    MARK_ALL_MESSAGES = False
    """
    Set this to True for :term:`legacy networks <Legacy Network>` that expects
    read marks for *all* messages and not just the latest one that was read
    (as most XMPP clients will only send a read mark for the latest msg).
    """

    PROPER_RECEIPTS = False
    """
    Set this to True if the legacy service provides a real equivalent of message delivery receipts
    (:xep:`0184`), meaning that there is an event thrown when the actual device of a contact receives
    a message. Make sure to call Contact.received() adequately if this is set to True.
    """

    GROUPS = False

    mtype: MessageTypes = "chat"
    is_group = False
    _can_send_carbon = False

    def __init__(self):
        self.datetime_started = datetime.now()
        self.xmpp = self  # ugly hack to work with the BaseSender mixin :/
        self.default_ns = "jabber:component:accept"
        super().__init__(
            config.JID,
            config.SECRET,
            config.SERVER,
            config.PORT,
            plugin_whitelist=SLIXMPP_PLUGINS,
            plugin_config={
                "xep_0077": {
                    "form_fields": None,
                    "form_instructions": self.REGISTRATION_INSTRUCTIONS,
                    "enable_subscription": self.REGISTRATION_TYPE
                    == RegistrationType.SINGLE_STEP_FORM,
                },
                "xep_0100": {
                    "component_name": self.COMPONENT_NAME,
                    "user_store": user_store,
                    "type": self.COMPONENT_TYPE,
                },
                "xep_0184": {
                    "auto_ack": False,
                    "auto_request": False,
                },
                "xep_0363": {
                    "upload_service": config.UPLOAD_SERVICE,
                },
            },
            fix_error_ns=True,
        )
        self.loop.set_exception_handler(self.__exception_handler)
        self.http: aiohttp.ClientSession = aiohttp.ClientSession()
        self.has_crashed: bool = False
        self.use_origin_id = False

        self.jid_validator: re.Pattern = re.compile(config.USER_JID_VALIDATOR)
        self.qr_pending_registrations = dict[str, asyncio.Future[bool]]()

        self.session_cls: BaseSession = BaseSession.get_unique_subclass()
        self.session_cls.xmpp = self

        self.get_session_from_stanza: Callable[
            [Union[Message, Presence, Iq]], BaseSession
        ] = self.session_cls.from_stanza  # type: ignore
        self.get_session_from_user: Callable[[GatewayUser], BaseSession] = (
            self.session_cls.from_user
        )

        self.register_plugins()
        self.__register_slixmpp_events()
        self.roster.set_backend(RosterBackend)

        self.register_plugin("pubsub", {"component_name": self.COMPONENT_NAME})
        self.pubsub: PubSubComponent = self["pubsub"]
        self.vcard: VCard4Provider = self["xep_0292_provider"]
        self.delivery_receipt: DeliveryReceipt = DeliveryReceipt(self)

        # with this we receive user avatar updates
        self.plugin["xep_0030"].add_feature("urn:xmpp:avatar:metadata+notify")

        if self.GROUPS:
            self.plugin["xep_0030"].add_feature("http://jabber.org/protocol/muc")
            self.plugin["xep_0030"].add_feature("urn:xmpp:mam:2")
            self.plugin["xep_0030"].add_feature("urn:xmpp:mam:2#extended")
            self.plugin["xep_0030"].add_feature(self.plugin["xep_0421"].namespace)
            self.plugin["xep_0030"].add_feature(self["xep_0317"].stanza.NS)
            self.plugin["xep_0030"].add_identity(
                category="conference",
                name=self.COMPONENT_NAME,
                itype="text",
                jid=self.boundjid,
            )

        # why does mypy need these type annotations? no idea
        self.__adhoc_handler: AdhocProvider = AdhocProvider(self)
        self.__chat_commands_handler: ChatCommandProvider = ChatCommandProvider(self)
        self.__disco_handler = Disco(self)

        self.__ping_handler = Ping(self)
        self.__mam_handler = Mam(self)
        self.__search_handler = Search(self)
        self.__caps_handler = Caps(self)
        self.__vcard_temp_handler = VCardTemp(self)
        self.__muc_admin_handler = MucAdmin(self)
        self.__registration = Registration(self)
        self.__dispatcher = SessionDispatcher(self)

        self.__register_commands()

        db.mam_launch_cleanup_task(self.loop)

    def __register_commands(self):
        for cls in Command.subclasses:
            if any(x is NotImplemented for x in [cls.CHAT_COMMAND, cls.NODE, cls.NAME]):
                log.debug("Not adding command '%s' because it looks abstract", cls)
                continue
            if cls is Exec:
                if config.DEV_MODE:
                    log.warning("/!\ DEV MODE ENABLED /!\\")
                else:
                    continue
            c = cls(self)
            log.debug("Registering %s", cls)
            self.__adhoc_handler.register(c)
            self.__chat_commands_handler.register(c)

    def __exception_handler(self, loop: asyncio.AbstractEventLoop, context):
        """
        Called when a task created by loop.create_task() raises an Exception

        :param loop:
        :param context:
        :return:
        """
        log.debug("Context in the exception handler: %s", context)
        exc = context.get("exception")
        if exc is None:
            log.warning("No exception in this context: %s", context)
        elif isinstance(exc, SystemExit):
            log.debug("SystemExit called in an asyncio task")
        else:
            log.error("Crash in an asyncio task: %s", context)
            log.exception("Crash in task", exc_info=exc)
            self.has_crashed = True
            loop.stop()

    def __register_slixmpp_events(self):
        self.add_event_handler("session_start", self.__on_session_start)
        self.add_event_handler("disconnected", self.connect)
        self.add_event_handler("user_register", self._on_user_register)
        self.add_event_handler("user_unregister", self._on_user_unregister)
        self.add_event_handler("groupchat_message_error", self.__on_group_chat_error)

    @property  # type: ignore
    def jid(self):
        # Override to avoid slixmpp deprecation warnings.
        return self.boundjid

    async def __on_group_chat_error(self, msg: Message):
        condition = msg["error"].get_condition()
        if condition not in KICKABLE_ERRORS:
            return

        try:
            muc = await self.get_muc_from_stanza(msg)
        except XMPPError as e:
            log.debug("Not removing resource", exc_info=e)
            return
        mfrom = msg.get_from()
        resource = mfrom.resource
        try:
            muc.user_resources.remove(resource)
        except KeyError:
            # this actually happens quite frequently on for both beagle and monal
            # (not sure why?), but is of no consequence
            log.debug("%s was not in the resources of %s", resource, muc)
        else:
            log.info(
                "Removed %s from the resources of %s because of error", resource, muc
            )

    async def __on_session_start(self, event):
        log.debug("Gateway session start: %s", event)

        # prevents XMPP clients from considering the gateway as an HTTP upload
        disco = self.plugin["xep_0030"]
        await disco.del_feature(feature="urn:xmpp:http:upload:0", jid=self.boundjid)
        await self.plugin["xep_0115"].update_caps(jid=self.boundjid)

        await self.pubsub.set_avatar(
            jid=self.boundjid.bare, avatar=self.COMPONENT_AVATAR
        )

        for user in user_store.get_all():
            # TODO: before this, we should check if the user has removed us from their roster
            #       while we were offline and trigger unregister from there. Presence probe does not seem
            #       to work in this case, there must be another way. privileged entity could be used
            #       as last resort.
            try:
                await self["xep_0100"].add_component_to_roster(user.jid)
                await self.__add_component_to_mds_whitelist(user.jid)
            except IqError as e:
                # TODO: remove the user when this happens? or at least
                # this can happen when the user has unsubscribed from the XMPP server
                log.warning(
                    "Error with user %s, not logging them automatically",
                    user,
                    exc_info=e,
                )
                continue
            self.send_presence(
                pto=user.bare_jid, ptype="probe"
            )  # ensure we get all resources for user
            session = self.session_cls.from_user(user)
            self.loop.create_task(self.__login_wrap(session))

        log.info("Slidge has successfully started")

    async def __add_component_to_mds_whitelist(self, user_jid: JID):
        # Uses privileged entity to add ourselves to the whitelist of the PEP
        # MDS node so we receive MDS events
        iq_creation = Iq(sto=user_jid.bare, sfrom=user_jid, stype="set")
        iq_creation["pubsub"]["create"]["node"] = self["xep_0490"].stanza.NS

        try:
            await self["xep_0356"].send_privileged_iq(iq_creation)
        except PermissionError:
            log.warning(
                "IQ privileges not granted for pubsub namespace, we cannot "
                "create the MDS node of %s",
                user_jid,
            )
        except IqError as e:
            # conflict this means the node already exists, we can ignore that
            if e.condition != "conflict":
                log.exception(
                    "Could not create the MDS node of %s", user_jid, exc_info=e
                )
        except Exception as e:
            log.exception(
                "Error while trying to create to the MDS node of %s",
                user_jid,
                exc_info=e,
            )

        iq_affiliation = Iq(sto=user_jid.bare, sfrom=user_jid, stype="set")
        iq_affiliation["pubsub_owner"]["affiliations"]["node"] = self[
            "xep_0490"
        ].stanza.NS

        aff = OwnerAffiliation()
        aff["jid"] = self.boundjid.bare
        aff["affiliation"] = "member"
        iq_affiliation["pubsub_owner"]["affiliations"].append(aff)

        try:
            await self["xep_0356"].send_privileged_iq(iq_affiliation)
        except PermissionError:
            log.warning(
                "IQ privileges not granted for pubsub#owner namespace, we cannot "
                "listen to the MDS events of %s",
                user_jid,
            )
        except Exception as e:
            log.exception(
                "Error while trying to subscribe to the MDS node of %s",
                user_jid,
                exc_info=e,
            )

    async def __login_wrap(self, session: "BaseSession"):
        session.send_gateway_status("Logging in…", show="dnd")
        try:
            status = await session.login()
        except Exception as e:
            log.warning("Login problem for %s", session.user, exc_info=e)
            log.exception(e)
            session.send_gateway_status(f"Could not login: {e}", show="busy")
            session.send_gateway_message(
                "You are not connected to this gateway! "
                f"Maybe this message will tell you why: {e}"
            )
            return

        log.info("Login success for %s", session.user)
        session.logged = True
        session.send_gateway_status("Syncing contacts…", show="dnd")
        await session.contacts.fill()
        if not (r := session.contacts.ready).done():
            r.set_result(True)
        if self.GROUPS:
            session.send_gateway_status("Syncing groups…", show="dnd")
            await session.bookmarks.fill()
            if not (r := session.bookmarks.ready).done():
                r.set_result(True)
        for c in session.contacts:
            # we need to receive presences directed at the contacts, in
            # order to send pubsub events for their +notify features
            self.send_presence(pfrom=c.jid, pto=session.user.bare_jid, ptype="probe")
        if status is None:
            session.send_gateway_status("Logged in", show="chat")
        else:
            session.send_gateway_status(status, show="chat")
        # If we stored users avatars (or their hash) persistently across slidge
        # restarts, we would not need to fetch it on startup
        self.loop.create_task(self.__fetch_user_avatar(session))

    async def __fetch_user_avatar(self, session: BaseSession):
        try:
            iq = await self.xmpp.plugin["xep_0060"].get_items(
                session.user.bare_jid,
                self.xmpp.plugin["xep_0084"].stanza.MetaData.namespace,
                ifrom=self.boundjid.bare,
            )
        except IqError as e:
            session.log.debug("Failed to retrieve avatar: %r", e)
            return
        await self.__dispatcher.on_avatar_metadata_info(
            session, iq["pubsub"]["items"]["item"]["avatar_metadata"]["info"]
        )

    def _send(
        self, stanza: MessageOrPresenceTypeVar, **send_kwargs
    ) -> MessageOrPresenceTypeVar:
        stanza.set_from(self.boundjid.bare)
        if mto := send_kwargs.get("mto"):
            stanza.set_to(mto)
        stanza.send()
        return stanza

    async def _on_user_register(self, iq: Iq):
        session = self.get_session_from_stanza(iq)
        for jid in config.ADMINS:
            self.send_message(
                mto=jid,
                mbody=f"{iq.get_from()} has registered",
                mtype="headline",
                mfrom=self.boundjid.bare,
            )
        session.send_gateway_message(self.WELCOME_MESSAGE)
        await self.__login_wrap(session)

    async def _on_user_unregister(self, iq: Iq):
        await self.session_cls.kill_by_jid(iq.get_from())

    def raise_if_not_allowed_jid(self, jid: JID):
        if not self.jid_validator.match(jid.bare):
            raise XMPPError(
                condition="not-allowed",
                text="Your account is not allowed to use this gateway.",
            )

    def send_raw(self, data: Union[str, bytes]):
        # overridden from XMLStream to strip base64-encoded data from the logs
        # to make them more readable.
        if log.isEnabledFor(level=logging.DEBUG):
            if isinstance(data, str):
                stripped = copy(data)
            else:
                stripped = data.decode("utf-8")
            # there is probably a way to do that in a single RE,
            # but since it's only for debugging, the perf penalty
            # does not matter much
            for el in LOG_STRIP_ELEMENTS:
                stripped = re.sub(
                    f"(<{el}.*?>)(.*)(</{el}>)",
                    "\1[STRIPPED]\3",
                    stripped,
                    flags=re.DOTALL | re.IGNORECASE,
                )
            log.debug("SEND: %s", stripped)
        if not self.transport:
            raise NotConnectedError()
        if isinstance(data, str):
            data = data.encode("utf-8")
        self.transport.write(data)

    def get_session_from_jid(self, j: JID):
        try:
            return self.session_cls.from_jid(j)
        except XMPPError:
            pass

    async def get_muc_from_stanza(self, iq: Union[Iq, Message]) -> "LegacyMUC":
        ito = iq.get_to()

        if ito == self.boundjid.bare:
            raise XMPPError(
                text="No MAM on the component itself, use a JID with a resource"
            )

        ifrom = iq.get_from()
        user = user_store.get_by_jid(ifrom)
        if user is None:
            raise XMPPError("registration-required")

        session = self.get_session_from_user(user)
        session.raise_if_not_logged()

        muc = await session.bookmarks.by_jid(ito)

        return muc

    def exception(self, exception: Exception):
        # """
        # Called when a task created by slixmpp's internal (eg, on slix events) raises an Exception.
        #
        # Stop the event loop and exit on unhandled exception.
        #
        # The default :class:`slixmpp.basexmpp.BaseXMPP` behaviour is just to
        # log the exception, but we want to avoid undefined behaviour.
        #
        # :param exception: An unhandled :class:`Exception` object.
        # """
        if isinstance(exception, IqError):
            iq = exception.iq
            log.error("%s: %s", iq["error"]["condition"], iq["error"]["text"])
            log.warning("You should catch IqError exceptions")
        elif isinstance(exception, IqTimeout):
            iq = exception.iq
            log.error("Request timed out: %s", iq)
            log.warning("You should catch IqTimeout exceptions")
        elif isinstance(exception, SyntaxError):
            # Hide stream parsing errors that occur when the
            # stream is disconnected (they've been handled, we
            # don't need to make a mess in the logs).
            pass
        else:
            if exception:
                log.exception(exception)
            self.loop.stop()
            exit(1)

    def re_login(self, session: "BaseSession"):
        async def w():
            await session.logout()
            await self.__login_wrap(session)

        self.loop.create_task(w())

    async def make_registration_form(self, _jid, _node, _ifrom, iq: Iq):
        self.raise_if_not_allowed_jid(iq.get_from())
        reg = iq["register"]
        user = user_store.get_by_stanza(iq)
        log.debug("User found: %s", user)

        form = reg["form"]
        form.add_field(
            "FORM_TYPE",
            ftype="hidden",
            value="jabber:iq:register",
        )
        form["title"] = f"Registration to '{self.COMPONENT_NAME}'"
        form["instructions"] = self.REGISTRATION_INSTRUCTIONS

        if user is not None:
            reg["registered"] = False
            form.add_field(
                "remove",
                label="Remove my registration",
                required=True,
                ftype="boolean",
                value=False,
            )

        for field in self.REGISTRATION_FIELDS:
            if field.var in reg.interfaces:
                val = None if user is None else user.get(field.var)
                if val is None:
                    reg.add_field(field.var)
                else:
                    reg[field.var] = val

        reg["instructions"] = self.REGISTRATION_INSTRUCTIONS

        for field in self.REGISTRATION_FIELDS:
            form.add_field(
                field.var,
                label=field.label,
                required=field.required,
                ftype=field.type,
                options=field.options,
                value=field.value if user is None else user.get(field.var, field.value),
            )

        reply = iq.reply()
        reply.set_payload(reg)
        return reply

    async def user_prevalidate(self, ifrom: JID, form_dict: dict[str, Optional[str]]):
        # Pre validate a registration form using the content of self.REGISTRATION_FIELDS
        # before passing it to the plugin custom validation logic.
        for field in self.REGISTRATION_FIELDS:
            if field.required and not form_dict.get(field.var):
                raise ValueError(f"Missing field: '{field.label}'")

        await self.validate(ifrom, form_dict)

    async def validate(
        self, user_jid: JID, registration_form: dict[str, Optional[str]]
    ):
        """
        Validate a user's initial registration form.

        Should raise the appropriate :class:`slixmpp.exceptions.XMPPError`
        if the registration does not allow to continue the registration process.

        If :py:attr:`REGISTRATION_TYPE` is a
        :attr:`.RegistrationType.SINGLE_STEP_FORM`,
        this method should raise something if it wasn't possible to successfully
        log in to the legacy service with the registration form content.

        It is also used for other types of :py:attr:`REGISTRATION_TYPE` too, since
        the first step is always a form. If :attr:`.REGISTRATION_FIELDS` is an
        empty list (ie, it declares no :class:`.FormField`), the "form" is
        effectively a confirmation dialog displaying
        :attr:`.REGISTRATION_INSTRUCTIONS`.

        :param user_jid: JID of the user that has just registered
        :param registration_form: A dict where keys are the :attr:`.FormField.var` attributes
         of the :attr:`.BaseGateway.REGISTRATION_FIELDS` iterable
        """
        raise NotImplementedError

    async def validate_two_factor_code(self, user: GatewayUser, code: str):
        """
        Called when the user enters their 2FA code.

        Should raise the appropriate :class:`slixmpp.exceptions.XMPPError`
        if the login fails, and return successfully otherwise.

        Only used when :attr:`REGISTRATION_TYPE` is
        :attr:`.RegistrationType.TWO_FACTOR_CODE`.

        :param user: The :class:`.GatewayUser` whose registration is pending
            Use their :attr:`.GatewayUser.bare_jid` and/or
            :attr:`.registration_form` attributes to get what you need.
        :param code: The code they entered, either via "chatbot" message or
            adhoc command
        """
        raise NotImplementedError

    async def get_qr_text(self, user: GatewayUser) -> str:
        """
        This is where slidge gets the QR code content for the QR-based
        registration process. It will turn it into a QR code image and send it
        to the not-yet-fully-registered :class:`.GatewayUser`.

        Only used in when :attr:`BaseGateway.REGISTRATION_TYPE` is
        :attr:`.RegistrationType.QRCODE`.

        :param user: The :class:`.GatewayUser` whose registration is pending
            Use their :attr:`.GatewayUser.bare_jid` and/or
            :attr:`.registration_form` attributes to get what you need.
        """
        raise NotImplementedError

    async def confirm_qr(
        self, user_bare_jid: str, exception: Optional[Exception] = None
    ):
        """
        This method is meant to be called to finalize QR code-based registration
        flows, once the legacy service confirms the QR flashing.

        Only used in when :attr:`BaseGateway.REGISTRATION_TYPE` is
        :attr:`.RegistrationType.QRCODE`.

        :param user_bare_jid: The bare JID of the almost-registered
            :class:`GatewayUser` instance
        :param exception: Optionally, an XMPPError to be raised to **not** confirm
            QR code flashing.
        """
        fut = self.qr_pending_registrations[user_bare_jid]
        if exception is None:
            fut.set_result(True)
        else:
            fut.set_exception(exception)

    async def unregister_user(self, user: GatewayUser):
        await self.xmpp.plugin["xep_0077"].api["user_remove"](None, None, user.jid)
        await self.xmpp.session_cls.kill_by_jid(user.jid)

    async def unregister(self, user: GatewayUser):
        """
        Optionally override this if you need to clean additional
        stuff after a user has been removed from the permanent user_store.

        By default, this just calls :meth:`BaseSession.logout`.

        :param user:
        """
        session = self.get_session_from_user(user)
        await session.logout()

    async def input(
        self, jid: JID, text=None, mtype: MessageTypes = "chat", **msg_kwargs
    ) -> str:
        """
        Request arbitrary user input using a simple chat message, and await the result.

        You shouldn't need to call this directly bust instead use
        :meth:`.BaseSession.input` to directly target a user.

        :param jid: The JID we want input from
        :param text: A prompt to display for the user
        :param mtype: Message type
        :return: The user's reply
        """
        return await self.__chat_commands_handler.input(jid, text, mtype, **msg_kwargs)

    async def send_qr(self, text: str, **msg_kwargs):
        """
        Sends a QR Code to a JID

        You shouldn't need to call directly bust instead use
        :meth:`.BaseSession.send_qr` to directly target a user.

        :param text: The text that will be converted to a QR Code
        :param msg_kwargs: Optional additional arguments to pass to
            :meth:`.BaseGateway.send_file`, such as the recipient of the QR,
            code
        """
        qr = qrcode.make(text)
        with tempfile.NamedTemporaryFile(
            suffix=".png", delete=config.NO_UPLOAD_METHOD != "move"
        ) as f:
            qr.save(f.name)
            await self.send_file(f.name, **msg_kwargs)

    def shutdown(self):
        # """
        # Called by the slidge entrypoint on normal exit.
        #
        # Sends offline presences from all contacts of all user sessions and from
        # the gateway component itself.
        # No need to call this manually, :func:`slidge.__main__.main` should take care of it.
        # """
        log.debug("Shutting down")
        for user in user_store.get_all():
            self.session_cls.from_jid(user.jid).shutdown()
            self.send_presence(ptype="unavailable", pto=user.jid)


KICKABLE_ERRORS = {
    "gone",
    "internal-server-error",
    "item-not-found",
    "jid-malformed",
    "recipient-unavailable",
    "redirect",
    "remote-server-not-found",
    "remote-server-timeout",
    "service-unavailable",
    "malformed error",
}


SLIXMPP_PLUGINS = [
    "link_preview",  # https://wiki.soprani.ca/CheogramApp/LinkPreviews
    "xep_0030",  # Service discovery
    "xep_0045",  # Multi-User Chat
    "xep_0050",  # Adhoc commands
    "xep_0054",  # VCard-temp (for MUC avatars)
    "xep_0055",  # Jabber search
    "xep_0059",  # Result Set Management
    "xep_0066",  # Out of Band Data
    "xep_0077",  # In-band registration
    "xep_0084",  # User Avatar
    "xep_0085",  # Chat state notifications
    "xep_0100",  # Gateway interaction
    "xep_0106",  # JID Escaping
    "xep_0115",  # Entity capabilities
    "xep_0122",  # Data Forms Validation
    "xep_0153",  # vCard-Based Avatars (for MUC avatars)
    "xep_0172",  # User nickname
    "xep_0184",  # Message Delivery Receipts
    "xep_0199",  # XMPP Ping
    "xep_0221",  # Data Forms Media Element
    "xep_0249",  # Direct MUC Invitations
    "xep_0264",  # Jingle Content Thumbnails
    "xep_0280",  # Carbons
    "xep_0292_provider",  # VCard4
    "xep_0308",  # Last message correction
    "xep_0313",  # Message Archive Management
    "xep_0317",  # Hats
    "xep_0319",  # Last User Interaction in Presence
    "xep_0333",  # Chat markers
    "xep_0334",  # Message Processing Hints
    "xep_0356",  # Privileged Entity
    "xep_0356_old",  # Privileged Entity (old namespace)
    "xep_0363",  # HTTP file upload
    "xep_0385",  # Stateless in-line media sharing
    "xep_0402",  # PEP Native Bookmarks
    "xep_0421",  # Anonymous unique occupant identifiers for MUCs
    "xep_0424",  # Message retraction
    "xep_0425",  # Message moderation
    "xep_0444",  # Message reactions
    "xep_0447",  # Stateless File Sharing
    "xep_0461",  # Message replies
    "xep_0490",  # Message Displayed Synchronization
]

LOG_STRIP_ELEMENTS = ["data", "binval"]

log = logging.getLogger(__name__)



from slixmpp.plugins.xep_0297 import Forwarded
from slixmpp.stanza import Message
from slixmpp.xmlstream import ElementBase, register_stanza_plugin


class PrivilegeOld(ElementBase):
    namespace = "urn:xmpp:privilege:1"
    name = "privilege"
    plugin_attrib = "privilege_old"

    def permission(self, access):
        for perm in self["perms"]:
            if perm["access"] == access:
                return perm["type"]

    def roster(self):
        return self.permission("roster")

    def message(self):
        return self.permission("message")

    def presence(self):
        return self.permission("presence")

    def add_perm(self, access, type):
        # This should only be needed for servers, so maybe out of scope for slixmpp
        perm = PermOld()
        perm["type"] = type
        perm["access"] = access
        self.append(perm)


class PermOld(ElementBase):
    namespace = "urn:xmpp:privilege:1"
    name = "perm"
    plugin_attrib = "perm"
    plugin_multi_attrib = "perms"
    interfaces = {"type", "access"}


def register():
    register_stanza_plugin(Message, PrivilegeOld)
    register_stanza_plugin(PrivilegeOld, Forwarded)
    register_stanza_plugin(PrivilegeOld, PermOld, iterable=True)



import logging
from copy import copy
from typing import TYPE_CHECKING, Awaitable, Callable, Optional, Union

from slixmpp import JID, CoroutineCallback, Iq, Message, Presence, StanzaPath
from slixmpp.exceptions import XMPPError
from slixmpp.plugins.xep_0004 import Form
from slixmpp.plugins.xep_0084.stanza import Info

from ... import LegacyContact
from ...group.room import LegacyMUC
from ...util.sql import db
from ...util.types import LinkPreview, Recipient, RecipientType
from ...util.util import (
    dict_to_named_tuple,
    merge_resources,
    remove_emoji_variation_selector_16,
)
from .. import config
from ..session import BaseSession

if TYPE_CHECKING:
    from .base import BaseGateway

HandlerType = Callable[[Union[Presence, Message]], Awaitable[None]]


class Ignore(BaseException):
    pass


class SessionDispatcher:
    def __init__(self, xmpp: "BaseGateway"):
        self.xmpp = xmpp
        self.http = xmpp.http

        xmpp.register_handler(
            CoroutineCallback(
                "MUCModerate",
                StanzaPath("iq/apply_to/moderate"),
                _exceptions_to_xmpp_errors(self.on_user_moderation),  # type:ignore
            )
        )
        xmpp.register_handler(
            CoroutineCallback(
                "MUCSetAffiliation",
                StanzaPath("iq@type=set/mucadmin_query"),
                _exceptions_to_xmpp_errors(self.on_user_set_affiliation),  # type:ignore
            )
        )
        xmpp.register_handler(
            CoroutineCallback(
                "muc#admin",
                StanzaPath("iq@type=get/mucowner_query"),
                _exceptions_to_xmpp_errors(self.on_muc_owner_query),  # type: ignore
            )
        )
        xmpp.register_handler(
            CoroutineCallback(
                "muc#admin",
                StanzaPath("iq@type=set/mucowner_query"),
                _exceptions_to_xmpp_errors(self.on_muc_owner_set),  # type: ignore
            )
        )

        for event in (
            "legacy_message",
            "marker_displayed",
            "presence",
            "chatstate_active",
            "chatstate_inactive",
            "chatstate_composing",
            "chatstate_paused",
            "message_correction",
            "reactions",
            "message_retract",
            "groupchat_join",
            "groupchat_message",
            "groupchat_direct_invite",
            "groupchat_subject",
            "avatar_metadata_publish",
            "message_displayed_synchronization_publish",
        ):
            xmpp.add_event_handler(
                event, _exceptions_to_xmpp_errors(getattr(self, "on_" + event))
            )

    async def __get_session(
        self, stanza: Union[Message, Presence, Iq], timeout: Optional[int] = 10
    ) -> BaseSession:
        xmpp = self.xmpp
        if stanza.get_from().server == xmpp.boundjid.bare:
            log.debug("Ignoring echo")
            raise Ignore
        if (
            isinstance(stanza, Message)
            and stanza.get_type() == "chat"
            and stanza.get_to() == xmpp.boundjid.bare
        ):
            log.debug("Ignoring message to component")
            raise Ignore
        session = xmpp.get_session_from_stanza(stanza)
        await session.wait_for_ready(timeout)
        if isinstance(stanza, Message) and _ignore(session, stanza):
            raise Ignore
        return session

    def __ack(self, msg: Message):
        if not self.xmpp.PROPER_RECEIPTS:
            self.xmpp.delivery_receipt.ack(msg)

    async def __get_session_entity_thread(
        self, msg: Message
    ) -> tuple["BaseSession", Recipient, Union[int, str]]:
        session = await self.__get_session(msg)
        e: Recipient = await _get_entity(session, msg)
        legacy_thread = await _xmpp_to_legacy_thread(session, msg, e)
        return session, e, legacy_thread

    async def on_legacy_message(self, msg: Message):
        """
        Meant to be called from :class:`BaseGateway` only.

        :param msg:
        :return:
        """
        # we MUST not use `if m["replace"]["id"]` because it adds the tag if not
        # present. this is a problem for MUC echoed messages
        if msg.get_plugin("replace", check=True) is not None:
            # ignore last message correction (handled by a specific method)
            return
        if msg.get_plugin("apply_to", check=True) is not None:
            # ignore message retraction (handled by a specific method)
            return
        if msg.get_plugin("reactions", check=True) is not None:
            # ignore message reaction fallback.
            # the reaction itself is handled by self.react_from_msg().
            return
        if msg.get_plugin("retract", check=True) is not None:
            # ignore message retraction fallback.
            # the retraction itself is handled by self.on_retract
            return

        session, entity, thread = await self.__get_session_entity_thread(msg)

        e: Recipient = await _get_entity(session, msg)
        log.debug("Entity %r", e)

        if msg.get_plugin("oob", check=True) is not None:
            url = msg["oob"]["url"]
        else:
            url = None

        text = msg["body"]

        reply_to = None
        reply_fallback = None
        if msg.get_plugin("reply", check=True):
            try:
                reply_to_msg_xmpp_id = _xmpp_msg_id_to_legacy(
                    session, msg["reply"]["id"]
                )
            except XMPPError:
                session.log.debug(
                    "Could not determine reply-to legacy msg ID, sending quote instead."
                )
                text = msg["body"]
                reply_fallback = None
                reply_to_msg_xmpp_id = None
            else:
                reply_to_jid = JID(msg["reply"]["to"])
                if msg["type"] == "chat":
                    if reply_to_jid.bare != session.user.jid.bare:
                        try:
                            reply_to = await session.contacts.by_jid(reply_to_jid)
                        except XMPPError:
                            pass
                elif msg["type"] == "groupchat":
                    nick = reply_to_jid.resource
                    try:
                        muc = await session.bookmarks.by_jid(reply_to_jid)
                    except XMPPError:
                        pass
                    else:
                        if nick != muc.user_nick:
                            reply_to = await muc.get_participant(
                                reply_to_jid.resource, store=False
                            )
                if msg.get_plugin("fallback", check=True) and (
                    isinstance(e, LegacyMUC) or e.REPLIES
                ):
                    text = msg["fallback"].get_stripped_body(
                        self.xmpp["xep_0461"].namespace
                    )
                    try:
                        reply_fallback = msg["reply"].get_fallback_body()
                    except AttributeError:
                        pass
        else:
            reply_to_msg_xmpp_id = None
            reply_to = None

        if msg.get_plugin("link_previews", check=True):
            pass

        kwargs = dict(
            reply_to_msg_id=reply_to_msg_xmpp_id,
            reply_to_fallback_text=reply_fallback,
            reply_to=reply_to,
            thread=thread,
        )

        if not url and isinstance(e, LegacyMUC):
            kwargs["mentions"] = await e.parse_mentions(text)

        if previews := msg["link_previews"]:
            kwargs["link_previews"] = [
                dict_to_named_tuple(p, LinkPreview) for p in previews
            ]

        if url:
            async with self.http.get(url) as response:
                if response.status >= 400:
                    session.log.warning(
                        (
                            "OOB url cannot be downloaded: %s, sending the URL as text"
                            " instead."
                        ),
                        response,
                    )
                    legacy_msg_id = await session.on_text(e, url, **kwargs)
                else:
                    legacy_msg_id = await session.on_file(
                        e, url, http_response=response, **kwargs
                    )
        elif text:
            legacy_msg_id = await session.on_text(e, text, **kwargs)
        else:
            log.debug("Ignoring %s", msg.get_id())
            return

        if isinstance(e, LegacyMUC):
            await e.echo(msg, legacy_msg_id)
            if legacy_msg_id is not None:
                session.muc_sent_msg_ids[legacy_msg_id] = msg.get_id()
        else:
            self.__ack(msg)
            if legacy_msg_id is not None:
                session.sent[legacy_msg_id] = msg.get_id()
                if session.MESSAGE_IDS_ARE_THREAD_IDS and (t := msg["thread"]):
                    session.threads[t] = legacy_msg_id

    async def on_groupchat_message(self, msg: Message):
        await self.on_legacy_message(msg)

    async def on_message_correction(self, msg: Message):
        session, entity, thread = await self.__get_session_entity_thread(msg)
        xmpp_id = msg["replace"]["id"]
        if isinstance(entity, LegacyMUC):
            legacy_id = session.muc_sent_msg_ids.inverse.get(xmpp_id)
        else:
            legacy_id = _xmpp_msg_id_to_legacy(session, xmpp_id)

        if isinstance(entity, LegacyMUC):
            mentions = await entity.parse_mentions(msg["body"])
        else:
            mentions = None

        if previews := msg["link_previews"]:
            link_previews = [dict_to_named_tuple(p, LinkPreview) for p in previews]
        else:
            link_previews = []

        if legacy_id is None:
            log.debug("Did not find legacy ID to correct")
            new_legacy_msg_id = await session.on_text(
                entity,
                "Correction:" + msg["body"],
                thread=thread,
                mentions=mentions,
                link_previews=link_previews,
            )
        elif (
            not msg["body"].strip()
            and config.CORRECTION_EMPTY_BODY_AS_RETRACTION
            and entity.RETRACTION
        ):
            await session.on_retract(entity, legacy_id, thread=thread)
            new_legacy_msg_id = None
        elif entity.CORRECTION:
            new_legacy_msg_id = await session.on_correct(
                entity,
                msg["body"],
                legacy_id,
                thread=thread,
                mentions=mentions,
                link_previews=link_previews,
            )
        else:
            session.send_gateway_message(
                "Last message correction is not supported by this legacy service. "
                "Slidge will send your correction as new message."
            )
            if (
                config.LAST_MESSAGE_CORRECTION_RETRACTION_WORKAROUND
                and entity.RETRACTION
                and legacy_id is not None
            ):
                if legacy_id is not None:
                    session.send_gateway_message(
                        "Slidge will attempt to retract the original message you wanted"
                        " to edit."
                    )
                    await session.on_retract(entity, legacy_id, thread=thread)

            new_legacy_msg_id = await session.on_text(
                entity,
                "Correction: " + msg["body"],
                thread=thread,
                mentions=mentions,
                link_previews=link_previews,
            )

        if isinstance(entity, LegacyMUC):
            if new_legacy_msg_id is not None:
                session.muc_sent_msg_ids[new_legacy_msg_id] = msg.get_id()
            await entity.echo(msg, new_legacy_msg_id)
        else:
            self.__ack(msg)
            if new_legacy_msg_id is not None:
                session.sent[new_legacy_msg_id] = msg.get_id()

    async def on_message_retract(self, msg: Message):
        session, entity, thread = await self.__get_session_entity_thread(msg)
        if not entity.RETRACTION:
            raise XMPPError(
                "bad-request",
                "This legacy service does not support message retraction.",
            )
        xmpp_id: str = msg["retract"]["id"]
        legacy_id = _xmpp_msg_id_to_legacy(session, xmpp_id)
        if legacy_id:
            await session.on_retract(entity, legacy_id, thread=thread)
            if isinstance(entity, LegacyMUC):
                await entity.echo(msg, None)
        else:
            log.debug("Ignored retraction from user")
        self.__ack(msg)

    async def on_marker_displayed(self, msg: Message):
        session = await self.__get_session(msg)

        e: Recipient = await _get_entity(session, msg)
        legacy_thread = await _xmpp_to_legacy_thread(session, msg, e)
        displayed_msg_id = msg["displayed"]["id"]
        if not isinstance(e, LegacyMUC) and self.xmpp.MARK_ALL_MESSAGES:
            to_mark = e.get_msg_xmpp_id_up_to(displayed_msg_id)  # type: ignore
            if to_mark is None:
                session.log.debug("Can't mark all messages up to %s", displayed_msg_id)
                to_mark = [displayed_msg_id]
        else:
            to_mark = [displayed_msg_id]
        for xmpp_id in to_mark:
            await session.on_displayed(
                e, _xmpp_msg_id_to_legacy(session, xmpp_id), legacy_thread
            )
            if isinstance(e, LegacyMUC):
                await e.echo(msg, None)

    async def on_chatstate_active(self, msg: Message):
        if msg["body"]:
            # if there is a body, it's handled in self.on_legacy_message()
            return
        session, entity, thread = await self.__get_session_entity_thread(msg)
        await session.on_active(entity, thread)

    async def on_chatstate_inactive(self, msg: Message):
        session, entity, thread = await self.__get_session_entity_thread(msg)
        await session.on_inactive(entity, thread)

    async def on_chatstate_composing(self, msg: Message):
        session, entity, thread = await self.__get_session_entity_thread(msg)
        await session.on_composing(entity, thread)

    async def on_chatstate_paused(self, msg: Message):
        session, entity, thread = await self.__get_session_entity_thread(msg)
        await session.on_paused(entity, thread)

    async def on_reactions(self, msg: Message):
        session, entity, thread = await self.__get_session_entity_thread(msg)
        react_to: str = msg["reactions"]["id"]

        special_msg = session.SPECIAL_MSG_ID_PREFIX and react_to.startswith(
            session.SPECIAL_MSG_ID_PREFIX
        )

        if special_msg:
            legacy_id = react_to
        else:
            legacy_id = _xmpp_msg_id_to_legacy(session, react_to)

        if not legacy_id:
            log.debug("Ignored reaction from user")
            raise XMPPError(
                "internal-server-error",
                "Could not convert the XMPP msg ID to a legacy ID",
            )

        emojis = [
            remove_emoji_variation_selector_16(r["value"]) for r in msg["reactions"]
        ]
        error_msg = None
        entity = entity

        if not special_msg:
            if entity.REACTIONS_SINGLE_EMOJI and len(emojis) > 1:
                error_msg = "Maximum 1 emoji/message"

            if not error_msg and (subset := await entity.available_emojis(legacy_id)):
                if not set(emojis).issubset(subset):
                    error_msg = f"You can only react with the following emojis: {''.join(subset)}"

        if error_msg:
            session.send_gateway_message(error_msg)
            if not isinstance(entity, LegacyMUC):
                # no need to carbon for groups, we just don't echo the stanza
                entity.react(legacy_id, carbon=True)  # type: ignore
            await session.on_react(entity, legacy_id, [], thread=thread)
            raise XMPPError("not-acceptable", text=error_msg)

        await session.on_react(entity, legacy_id, emojis, thread=thread)
        if isinstance(entity, LegacyMUC):
            await entity.echo(msg, None)
        else:
            self.__ack(msg)

        multi = db.attachment_get_associated_xmpp_ids(react_to)
        if not multi:
            return

        if isinstance(entity, LegacyMUC):
            for xmpp_id in multi:
                mc = copy(msg)
                mc["reactions"]["id"] = xmpp_id
                await entity.echo(mc)
        elif isinstance(entity, LegacyContact):
            for xmpp_id in multi:
                entity.react(legacy_id, emojis, xmpp_id=xmpp_id, carbon=True)

    async def on_presence(self, p: Presence):
        session = await self.__get_session(p)

        if p.get_to() != self.xmpp.boundjid.bare:
            return
        # NB: get_type() returns either a proper presence type or
        #     a presence show if available. Weird, weird, weird slix.
        if (ptype := p.get_type()) not in _USEFUL_PRESENCES:
            return
        resources = self.xmpp.roster[self.xmpp.boundjid.bare][p.get_from()].resources
        session.log.debug("Received a presence from %s", p.get_from())
        await session.on_presence(
            p.get_from().resource,
            ptype,  # type: ignore
            p["status"],
            resources,
            merge_resources(resources),
        )

    async def on_groupchat_join(self, p: Presence):
        if not self.xmpp.GROUPS:
            raise XMPPError(
                "feature-not-implemented",
                "This gateway does not implement multi-user chats.",
            )
        session = await self.__get_session(p)
        session.raise_if_not_logged()
        muc = await session.bookmarks.by_jid(p.get_to())
        await muc.join(p)

    async def on_message_displayed_synchronization_publish(self, msg: Message):
        session = await self.__get_session(msg, timeout=None)

        chat_jid = msg["pubsub_event"]["items"]["item"]["id"]
        chat = await session.get_contact_or_group_or_participant(JID(chat_jid))
        if not isinstance(chat, LegacyMUC):
            session.log.debug("Ignoring non-groupchat MDS event")
            return

        stanza_id = msg["pubsub_event"]["items"]["item"]["displayed"]["stanza_id"]["id"]
        await session.on_displayed(chat, _xmpp_msg_id_to_legacy(session, stanza_id))

    async def on_avatar_metadata_publish(self, m: Message):
        if not config.SYNC_AVATAR:
            return

        session = await self.__get_session(m, timeout=None)
        info = m["pubsub_event"]["items"]["item"]["avatar_metadata"]["info"]

        await self.on_avatar_metadata_info(session, info)

    async def on_avatar_metadata_info(self, session: BaseSession, info: Info):
        session.log.debug("Avatar metadata info: %s", info)
        hash_ = info["id"]

        if session.avatar_hash == hash_:
            return
        session.avatar_hash = hash_

        if hash_:
            iq = await self.xmpp.plugin["xep_0084"].retrieve_avatar(
                session.user.jid, hash_, ifrom=self.xmpp.boundjid.bare
            )
            bytes_ = iq["pubsub"]["items"]["item"]["avatar_data"]["value"]
            type_ = info["type"]
            height = info["height"]
            width = info["width"]
        else:
            bytes_ = type_ = height = width = hash_ = None
        try:
            await session.on_avatar(bytes_, hash_, type_, width, height)
        except NotImplementedError:
            pass
        except Exception as e:
            # If something goes wrong here, replying an error stanza will to the
            # avatar update will likely not show in most clients, so let's send
            # a normal message from the component to the user.
            session.send_gateway_message(
                f"Something went wrong trying to set your avatar: {e!r}"
            )

    async def on_user_moderation(self, iq: Iq):
        session = await self.__get_session(iq)
        session.raise_if_not_logged()

        muc = await session.bookmarks.by_jid(iq.get_to())

        apply_to = iq["apply_to"]
        xmpp_id = apply_to["id"]
        if not xmpp_id:
            raise XMPPError("bad-request", "Missing moderated message ID")

        moderate = apply_to["moderate"]
        if not moderate["retract"]:
            raise XMPPError(
                "feature-not-implemented",
                "Slidge only implements moderation/retraction",
            )

        legacy_id = _xmpp_msg_id_to_legacy(session, xmpp_id)
        await session.on_moderate(muc, legacy_id, moderate["reason"] or None)
        iq.reply(clear=True).send()

    async def on_user_set_affiliation(self, iq: Iq):
        session = await self.__get_session(iq)
        session.raise_if_not_logged()

        item = iq["mucadmin_query"]["item"]
        contact = await session.contacts.by_jid(JID(item["jid"]))

        muc = await session.bookmarks.by_jid(iq.get_to())

        await muc.on_set_affiliation(
            contact, item["affiliation"], item["reason"] or None, item["nick"] or None
        )
        iq.reply(clear=True).send()

    async def on_groupchat_direct_invite(self, msg: Message):
        session = await self.__get_session(msg)
        session.raise_if_not_logged()

        invite = msg["groupchat_invite"]
        jid = JID(invite["jid"])

        if jid.domain != self.xmpp.boundjid.bare:
            raise XMPPError(
                "bad-request",
                "Legacy contacts can only be invited to legacy groups, not standard XMPP MUCs.",
            )

        if invite["password"]:
            raise XMPPError(
                "bad-request", "Password-protected groups are not supported"
            )

        contact = await session.contacts.by_jid(msg.get_to())
        muc = await session.bookmarks.by_jid(jid)

        await session.on_invitation(contact, muc, invite["reason"] or None)

    async def on_muc_owner_query(self, iq: Iq):
        session = await self.__get_session(iq)
        session.raise_if_not_logged()

        muc = await session.bookmarks.by_jid(iq.get_to())

        reply = iq.reply()

        form = Form(title="Slidge room configuration")
        form["instructions"] = (
            "Complete this form to modify the configuration of your room."
        )
        form.add_field(
            var="FORM_TYPE",
            type="hidden",
            value="http://jabber.org/protocol/muc#roomconfig",
        )
        form.add_field(
            var="muc#roomconfig_roomname",
            label="Natural-Language Room Name",
            type="text-single",
            value=muc.name,
        )
        if muc.HAS_DESCRIPTION:
            form.add_field(
                var="muc#roomconfig_roomdesc",
                label="Short Description of Room",
                type="text-single",
                value=muc.description,
            )

        muc_owner = iq["mucowner_query"]
        muc_owner.append(form)
        reply.append(muc_owner)
        reply.send()

    async def on_muc_owner_set(self, iq: Iq):
        session = await self.__get_session(iq)
        session.raise_if_not_logged()
        muc = await session.bookmarks.by_jid(iq.get_to())
        query = iq["mucowner_query"]

        if form := query.get_plugin("form", check=True):
            values = form.get_values()
            await muc.on_set_config(
                name=values.get("muc#roomconfig_roomname"),
                description=(
                    values.get("muc#roomconfig_roomdesc")
                    if muc.HAS_DESCRIPTION
                    else None
                ),
            )
            form["type"] = "result"
            clear = False
        elif destroy := query.get_plugin("destroy", check=True):
            reason = destroy["reason"] or None
            await muc.on_destroy_request(reason)
            user_participant = await muc.get_user_participant()
            user_participant._affiliation = "none"
            user_participant._role = "none"
            presence = user_participant._make_presence(ptype="unavailable", force=True)
            presence["muc"].enable("destroy")
            if reason is not None:
                presence["muc"]["destroy"]["reason"] = reason
            user_participant._send(presence)
            session.bookmarks.remove(muc)
            clear = True
        else:
            raise XMPPError("bad-request")

        iq.reply(clear=clear).send()

    async def on_groupchat_subject(self, msg: Message):
        session = await self.__get_session(msg)
        session.raise_if_not_logged()
        muc = await session.bookmarks.by_jid(msg.get_to())
        if not muc.HAS_SUBJECT:
            raise XMPPError(
                "bad-request",
                "There are no room subject in here. "
                "Use the room configuration to update its name or description",
            )
        await muc.on_set_subject(msg["subject"])


def _xmpp_msg_id_to_legacy(session: "BaseSession", xmpp_id: str):
    sent = session.sent.inverse.get(xmpp_id)
    if sent:
        return sent

    multi = db.attachment_get_legacy_id_for_xmpp_id(xmpp_id)
    if multi:
        return multi

    try:
        return session.xmpp_to_legacy_msg_id(xmpp_id)
    except XMPPError:
        raise
    except Exception as e:
        log.debug("Couldn't convert xmpp msg ID to legacy ID.", exc_info=e)
        raise XMPPError(
            "internal-server-error", "Couldn't convert xmpp msg ID to legacy ID."
        )


def _ignore(session: "BaseSession", msg: Message):
    if (i := msg.get_id()) not in session.ignore_messages:
        return False
    session.log.debug("Ignored sent carbon: %s", i)
    session.ignore_messages.remove(i)
    return True


async def _xmpp_to_legacy_thread(
    session: "BaseSession", msg: Message, recipient: RecipientType
):
    xmpp_thread = msg["thread"]
    if not xmpp_thread:
        return

    if session.MESSAGE_IDS_ARE_THREAD_IDS:
        return session.threads.get(xmpp_thread)

    async with session.thread_creation_lock:
        legacy_thread = session.threads.get(xmpp_thread)
        if legacy_thread is None:
            legacy_thread = await recipient.create_thread(xmpp_thread)
            session.threads[xmpp_thread] = legacy_thread
    return legacy_thread


async def _get_entity(session: "BaseSession", m: Message) -> RecipientType:
    session.raise_if_not_logged()
    if m.get_type() == "groupchat":
        muc = await session.bookmarks.by_jid(m.get_to())
        r = m.get_from().resource
        if r not in muc.user_resources:
            session.xmpp.loop.create_task(muc.kick_resource(r))
            raise XMPPError("not-acceptable", "You are not connected to this chat")
        return muc
    else:
        return await session.contacts.by_jid(m.get_to())


def _exceptions_to_xmpp_errors(cb: HandlerType) -> HandlerType:
    async def wrapped(stanza: Union[Presence, Message]):
        try:
            await cb(stanza)
        except Ignore:
            pass
        except XMPPError:
            raise
        except NotImplementedError:
            log.debug("NotImplementedError raised in %s", cb)
            raise XMPPError(
                "feature-not-implemented", "Not implemented by the legacy module"
            )
        except Exception as e:
            log.error("Failed to handle incoming stanza: %s", stanza, exc_info=e)
            raise XMPPError("internal-server-error", str(e))

    return wrapped


_USEFUL_PRESENCES = {"available", "unavailable", "away", "chat", "dnd", "xa"}


log = logging.getLogger(__name__)



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging

from slixmpp.plugins.base import BasePlugin
from slixmpp.stanza import Presence
from slixmpp.xmlstream import register_stanza_plugin

from . import VCardTempUpdate, stanza

log = logging.getLogger(__name__)


class XEP_0153(BasePlugin):
    name = "xep_0153"
    description = "XEP-0153: vCard-Based Avatars (slidge, just for MUCs)"
    dependencies = {"xep_0054"}
    stanza = stanza

    def plugin_init(self):
        register_stanza_plugin(Presence, VCardTempUpdate)



from slixmpp import Presence
from slixmpp.xmlstream import ElementBase, register_stanza_plugin

NS = "urn:xmpp:hats:0"


class Hats(ElementBase):
    name = plugin_attrib = "hats"
    namespace = NS

    def add_hats(self, data: list[tuple[str, str]]):
        for uri, title in data:
            hat = Hat()
            hat["uri"] = uri
            hat["title"] = title
            self.append(hat)


class Hat(ElementBase):
    name = plugin_attrib = "hat"
    namespace = NS
    interfaces = {"uri", "title"}
    plugin_multi_attrib = "hats"


def register():
    register_stanza_plugin(Hats, Hat, iterable=True)
    register_stanza_plugin(Presence, Hats)



from datetime import timedelta
from pathlib import Path
from typing import Optional

from slixmpp import JID as JIDType


class _TimedeltaSeconds(timedelta):
    def __new__(cls, s: str):
        return super().__new__(cls, seconds=int(s))


# REQUIRED, so not default value

LEGACY_MODULE: str
LEGACY_MODULE__DOC = (
    "Importable python module containing (at least) "
    "a BaseGateway and a LegacySession subclass"
)

SERVER: str = "localhost"
SERVER__DOC = (
    "The XMPP server's host name. Defaults to localhost, which is the "
    "standard way of running slidge, on the same host as the XMPP server. "
    "The 'Jabber Component Protocol' (XEP-0114) does not mention encryption, "
    "so you *should* provide encryption another way, eg via port forwarding, if "
    "you change this."
)
SERVER__SHORT = "s"

SECRET: str
SECRET__DOC = "The gateway component's secret (required to connect to the XMPP server)"

JID: JIDType
JID__DOC = "The gateway component's JID"
JID__SHORT = "j"

PORT: str = "5347"
PORT__DOC = "The XMPP server's port for incoming component connections"
PORT__SHORT = "p"

# Dynamic default (depends on other values)

HOME_DIR: Path
HOME_DIR__DOC = (
    "Shelve file used to store persistent user data. "
    "Defaults to /var/lib/slidge/${SLIDGE_JID}. "
)
HOME_DIR__DYNAMIC_DEFAULT = True

USER_JID_VALIDATOR: str
USER_JID_VALIDATOR__DOC = (
    "Regular expression to restrict users that can register to the gateway, by JID. "
    "Defaults to .*@${SLIDGE_SERVER}, but since SLIDGE_SERVER is usually localhost, "
    "you probably want to change that to .*@example.com"
)
USER_JID_VALIDATOR__DYNAMIC_DEFAULT = True

# Optional, so default value + type hint if default is None

ADMINS: tuple[JIDType, ...] = ()
ADMINS__DOC = "JIDs of the gateway admins"


UPLOAD_SERVICE: Optional[str] = None
UPLOAD_SERVICE__DOC = (
    "JID of an HTTP upload service the gateway can use. "
    "This is optional, as it should be automatically determined via service"
    "discovery."
)

SECRET_KEY: Optional[str] = None
SECRET_KEY__DOC = "Encryption for disk storage"

NO_ROSTER_PUSH = False
NO_ROSTER_PUSH__DOC = "Do not fill users' rosters with legacy contacts automatically"

ROSTER_PUSH_PRESENCE_SUBSCRIPTION_REQUEST_FALLBACK = True
ROSTER_PUSH_PRESENCE_SUBSCRIPTION_REQUEST_FALLBACK__DOC = (
    "If True, legacy contacts will send a presence request subscription "
    "when privileged roster push does not work, eg, if XEP-0356 (privileged "
    "entity) is not available for the component."
)

AVATAR_SIZE = 200
AVATAR_SIZE__DOC = (
    "Maximum image size (width and height), image ratio will be preserved"
)

USE_ATTACHMENT_ORIGINAL_URLS = False
USE_ATTACHMENT_ORIGINAL_URLS__DOC = (
    "For legacy plugins in which attachments are publicly downloadable URLs, "
    "let XMPP clients directly download them from this URL. Note that this will "
    "probably leak your client IP to the legacy network."
)

UPLOAD_REQUESTER: Optional[str] = None
UPLOAD_REQUESTER__DOC = (
    "Set which JID should request the upload slots. Defaults to the component JID."
)

NO_UPLOAD_PATH: Optional[str] = None
NO_UPLOAD_PATH__DOC = (
    "Instead of using the XMPP server's HTTP upload component, copy files to this dir. "
    "You need to set NO_UPLOAD_URL_PREFIX too if you use this option, and configure "
    "an web server to serve files in this dir."
)

NO_UPLOAD_URL_PREFIX: Optional[str] = None
NO_UPLOAD_URL_PREFIX__DOC = (
    "Base URL that servers files in the dir set in the no-upload-path option, "
    "eg https://example.com:666/slidge-attachments/"
)

NO_UPLOAD_METHOD: str = "copy"
NO_UPLOAD_METHOD__DOC = (
    "Whether to 'copy', 'move', 'hardlink' or 'symlink' the files in no-upload-path."
)

NO_UPLOAD_FILE_READ_OTHERS = False
NO_UPLOAD_FILE_READ_OTHERS__DOC = (
    "After writing a file in NO_UPLOAD_PATH, change its permission so that 'others' can"
    " read it."
)

IGNORE_DELAY_THRESHOLD = _TimedeltaSeconds("300")
IGNORE_DELAY_THRESHOLD__DOC = (
    "Threshold, in seconds, below which the <delay> information is stripped "
    "out of emitted stanzas."
)

PARTIAL_REGISTRATION_TIMEOUT = 3600
PARTIAL_REGISTRATION_TIMEOUT__DOC = (
    "Timeout before registration and login. Only useful for legacy networks where "
    "a single step registration process is not enough."
)

LAST_SEEN_FALLBACK = True
LAST_SEEN_FALLBACK__DOC = (
    "When using XEP-0319 (Last User Interaction in Presence), use the presence status"
    " to display the last seen information in the presence status. Useful for clients"
    " that do not implement XEP-0319."
)

QR_TIMEOUT = 60
QR_TIMEOUT__DOC = "Timeout for QR code flashing confirmation."

DOWNLOAD_CHUNK_SIZE = 1024
DOWNLOAD_CHUNK_SIZE__DOC = "Chunk size when slidge needs to download files using HTTP."

LAST_MESSAGE_CORRECTION_RETRACTION_WORKAROUND = False
LAST_MESSAGE_CORRECTION_RETRACTION_WORKAROUND__DOC = (
    "If the legacy service does not support last message correction but supports"
    " message retractions, slidge can 'retract' the edited message when you edit from"
    " an XMPP client, as a workaround. This may only work for editing messages"
    " **once**. If the legacy service does not support retractions and this is set to"
    " true, when XMPP clients attempt to correct, this will send a new message."
)

FIX_FILENAME_SUFFIX_MIME_TYPE = False
FIX_FILENAME_SUFFIX_MIME_TYPE__DOC = (
    "Fix the Filename suffix based on the Mime Type of the file. Some clients (eg"
    " Conversations) may not inline files that have a wrong suffix for the MIME Type."
    " Therefore the MIME Type of the file is checked, if the suffix is not valid for"
    " that MIME Type, a valid one will be picked."
)

LOG_FILE: Optional[Path] = None
LOG_FILE__DOC = "Log to a file instead of stdout/err"

MAM_MAX_DAYS = 7
MAM_MAX_DAYS__DOC = (
    "Maximum number of days for group archive retention. "
    "Since all text content stored in RAM right now, "
)

CORRECTION_EMPTY_BODY_AS_RETRACTION = True
CORRECTION_EMPTY_BODY_AS_RETRACTION__DOC = (
    "Treat last message correction to empty message as a retraction. "
    "(this is what cheogram do for retraction)"
)

ATTACHMENT_MAXIMUM_FILE_NAME_LENGTH = 200
ATTACHMENT_MAXIMUM_FILE_NAME_LENGTH__DOC = (
    "Some legacy network provide ridiculously long filenames, strip above this limit, "
    "preserving suffix."
)

ALWAYS_INVITE_WHEN_ADDING_BOOKMARKS = True
ALWAYS_INVITE_WHEN_ADDING_BOOKMARKS__DOC = (
    "Send an invitation to join MUCs when adding them to the bookmarks. While this "
    "should not be necessary, it helps with clients that do not support :xep:`0402` "
    "or that do not respect the auto-join flag."
)

AVATAR_RESAMPLING_THREADS = 2
AVATAR_RESAMPLING_THREADS__DOC = (
    "Number of additional threads to use for avatar resampling. Even in a single-core "
    "context, this makes avatar resampling non-blocking."
)

DEV_MODE = False
DEV_MODE__DOC = (
    "Enables an interactive python shell via chat commands, for admins."
    "Not safe to use in prod, but great during dev."
)

SYNC_AVATAR = True
SYNC_AVATAR__DOC = "Sync the user XMPP avatar to legacy network (if supported)."



import logging
import os
import sqlite3
import tempfile
from asyncio import AbstractEventLoop, Task, sleep
from datetime import datetime, timezone
from functools import lru_cache
from pathlib import Path
from time import time
from typing import (
    TYPE_CHECKING,
    Collection,
    Generic,
    Iterator,
    NamedTuple,
    Optional,
    TypeVar,
    Union,
)

from slixmpp import JID
from slixmpp.exceptions import XMPPError
from slixmpp.types import PresenceShows, PresenceTypes

from ..core import config
from .archive_msg import HistoryMessage

if TYPE_CHECKING:
    from .db import GatewayUser

KeyType = TypeVar("KeyType")
ValueType = TypeVar("ValueType")


class CachedPresence(NamedTuple):
    last_seen: Optional[datetime] = None
    ptype: Optional[PresenceTypes] = None
    pstatus: Optional[str] = None
    pshow: Optional[PresenceShows] = None


class MamMetadata(NamedTuple):
    id: str
    sent_on: datetime


class Base:
    def __init__(self):
        handler, filename = tempfile.mkstemp()

        os.close(handler)
        self.__filename = filename

        self.con = sqlite3.connect(filename)
        self.cur = self.con.cursor()
        self.cur.executescript((Path(__file__).parent / "schema.sql").read_text())

        self.__mam_cleanup_task: Optional[Task] = None

    def __del__(self):
        self.con.close()
        os.unlink(self.__filename)


class MAMMixin(Base):
    def __init__(self):
        super().__init__()
        self.__mam_cleanup_task: Optional[Task] = None
        self.__msg_cur = msg_cur = self.con.cursor()
        msg_cur.row_factory = self.__msg_factory  # type:ignore
        self.__metadata_cur = metadata_cur = self.con.cursor()
        metadata_cur.row_factory = self.__metadata_factory  # type:ignore

    @staticmethod
    def __msg_factory(_cur, row: tuple[str, float]) -> HistoryMessage:
        return HistoryMessage(
            row[0], when=datetime.fromtimestamp(row[1], tz=timezone.utc)
        )

    @staticmethod
    def __metadata_factory(_cur, row: tuple[str, float]) -> MamMetadata:
        return MamMetadata(row[0], datetime.fromtimestamp(row[1], tz=timezone.utc))

    def mam_nuke(self):
        self.cur.execute("DELETE FROM mam_message")
        self.con.commit()

    def mam_add_muc(self, jid: str, user: "GatewayUser"):
        try:
            self.cur.execute(
                "INSERT INTO "
                "muc(jid, user_id) "
                "VALUES("
                "  ?, "
                "  (SELECT id FROM user WHERE jid = ?)"
                ")",
                (jid, user.bare_jid),
            )
        except sqlite3.IntegrityError:
            log.debug("Tried to add a MUC that was already here: (%s, %s)", user, jid)
        else:
            self.con.commit()

    def mam_add_msg(self, muc_jid: str, msg: "HistoryMessage", user: "GatewayUser"):
        self.cur.execute(
            "REPLACE INTO "
            "mam_message(message_id, sender_jid, sent_on, xml, muc_id, user_id)"
            "VALUES(?, ?, ?, ?,"
            "(SELECT id FROM muc WHERE jid = ?),"
            "(SELECT id FROM user WHERE jid = ?)"
            ")",
            (
                msg.id,
                str(msg.stanza.get_from()),
                msg.when.timestamp(),
                str(msg.stanza),
                muc_jid,
                user.bare_jid,
            ),
        )
        self.con.commit()

    def mam_launch_cleanup_task(self, loop: AbstractEventLoop):
        self.__mam_cleanup_task = loop.create_task(self.__mam_cleanup())

    async def __mam_cleanup(self):
        await sleep(6 * 3600)
        self.mam_cleanup()

    def mam_cleanup(self):
        self.cur.execute(
            "DELETE FROM mam_message WHERE sent_on < ?",
            (time() - config.MAM_MAX_DAYS * 24 * 3600,),
        )
        self.con.commit()

    def __mam_get_sent_on(self, muc_jid: str, mid: str, user: "GatewayUser"):
        res = self.cur.execute(
            "SELECT sent_on "
            "FROM mam_message "
            "WHERE message_id = ? "
            "AND muc_id = (SELECT id FROM muc WHERE jid = ?) "
            "AND user_id = (SELECT id FROM user WHERE jid = ?)",
            (mid, muc_jid, user.bare_jid),
        )
        row = res.fetchone()
        if row is None:
            raise XMPPError("item-not-found", f"Message {mid} not found")
        return row[0]

    def __mam_bound(
        self,
        muc_jid: str,
        user: "GatewayUser",
        date: Optional[datetime] = None,
        id_: Optional[str] = None,
        comparator=min,
    ):
        if id_ is not None:
            after_id_sent_on = self.__mam_get_sent_on(muc_jid, id_, user)
            if date:
                timestamp = comparator(after_id_sent_on, date.timestamp())
            else:
                timestamp = after_id_sent_on
            return " AND sent_on > ?", timestamp
        elif date is None:
            raise TypeError
        else:
            return " AND sent_on >= ?", date.timestamp()

    def mam_get_messages(
        self,
        user: "GatewayUser",
        muc_jid: str,
        start_date: Optional[datetime] = None,
        end_date: Optional[datetime] = None,
        before_id: Optional[str] = None,
        after_id: Optional[str] = None,
        ids: Collection[str] = (),
        last_page_n: Optional[int] = None,
        sender: Optional[str] = None,
        flip=False,
    ) -> Iterator[HistoryMessage]:
        query = (
            "SELECT xml, sent_on FROM mam_message "
            "WHERE muc_id = (SELECT id FROM muc WHERE jid = ?) "
            "AND user_id = (SELECT id FROM user WHERE jid = ?) "
        )
        params: list[Union[str, float, int]] = [muc_jid, user.bare_jid]

        if start_date or after_id:
            subquery, timestamp = self.__mam_bound(
                muc_jid, user, start_date, after_id, max
            )
            query += subquery
            params.append(timestamp)
        if end_date or before_id:
            subquery, timestamp = self.__mam_bound(
                muc_jid, user, end_date, before_id, min
            )
            query += subquery
            params.append(timestamp)
        if sender:
            query += " AND sender_jid = ?"
            params.append(sender)
        if ids:
            query += f" AND message_id IN ({','.join('?' * len(ids))})"
            params.extend(ids)
        if last_page_n:
            # TODO: optimize query further when <flip> and last page are
            #       combined.
            query = f"SELECT * FROM ({query} ORDER BY sent_on DESC LIMIT ?)"
            params.append(last_page_n)
        query += " ORDER BY sent_on"
        if flip:
            query += " DESC"

        res = self.__msg_cur.execute(query, params)

        if ids:
            rows = res.fetchall()
            if len(rows) != len(ids):
                raise XMPPError(
                    "item-not-found",
                    "One of the requested messages IDs could not be found "
                    "with the given constraints.",
                )
            for row in rows:
                yield row

        while row := res.fetchone():
            yield row

    def mam_get_first_and_last(self, muc_jid: str) -> list[MamMetadata]:
        res = self.__metadata_cur.execute(
            "SELECT message_id, sent_on "
            "FROM mam_message "
            "JOIN muc ON muc.jid = ? "
            "WHERE sent_on = (SELECT MIN(sent_on) FROM mam_message WHERE muc_id = muc.id) "
            "   OR sent_on = (SELECT MAX(sent_on) FROM mam_message WHERE muc_id = muc.id) "
            " ORDER BY sent_on",
            (muc_jid,),
        )
        return res.fetchall()


class AttachmentMixin(Base):
    def attachment_remove(self, legacy_id):
        self.cur.execute("DELETE FROM attachment WHERE legacy_id = ?", (legacy_id,))
        self.con.commit()

    def attachment_store_url(self, legacy_id, url: str):
        self.cur.execute(
            "REPLACE INTO attachment(legacy_id, url) VALUES (?,?)", (legacy_id, url)
        )
        self.con.commit()

    def attachment_store_sims(self, url: str, sims: str):
        self.cur.execute("UPDATE attachment SET sims = ? WHERE url = ?", (sims, url))
        self.con.commit()

    def attachment_store_sfs(self, url: str, sfs: str):
        self.cur.execute("UPDATE attachment SET sfs = ? WHERE url = ?", (sfs, url))
        self.con.commit()

    def attachment_get_url(self, legacy_id):
        res = self.cur.execute(
            "SELECT url FROM attachment WHERE legacy_id = ?", (legacy_id,)
        )
        return first_of_tuple_or_none(res.fetchone())

    def attachment_get_sims(self, url: str):
        res = self.cur.execute("SELECT sims FROM attachment WHERE url = ?", (url,))
        return first_of_tuple_or_none(res.fetchone())

    def attachment_get_sfs(self, url: str):
        res = self.cur.execute("SELECT sfs FROM attachment WHERE url = ?", (url,))
        return first_of_tuple_or_none(res.fetchone())

    def attachment_store_legacy_to_multi_xmpp_msg_ids(
        self, legacy_id, xmpp_ids: list[str]
    ):
        with self.con:
            res = self.cur.execute(
                "INSERT OR IGNORE INTO attachment_legacy_msg_id(legacy_id) VALUES (?)",
                (legacy_id,),
            )
            row_id = res.lastrowid
            # for xmpp_id in xmpp_ids:
            self.cur.executemany(
                "INSERT INTO attachment_xmpp_ids(legacy_msg_id, xmpp_id) VALUES (?, ?)",
                ((row_id, i) for i in xmpp_ids),
            )

    def attachment_get_xmpp_ids_for_legacy_msg_id(self, legacy_id) -> list:
        res = self.cur.execute(
            "SELECT xmpp_id FROM attachment_xmpp_ids "
            "WHERE legacy_msg_id = (SELECT id FROM attachment_legacy_msg_id WHERE legacy_id = ?)",
            (legacy_id,),
        )
        return [r[0] for r in res.fetchall()]

    def attachment_get_associated_xmpp_ids(self, xmpp_id: str):
        res = self.cur.execute(
            "SELECT xmpp_id FROM attachment_xmpp_ids "
            "WHERE legacy_msg_id = "
            "(SELECT legacy_msg_id FROM attachment_xmpp_ids WHERE xmpp_id = ?)",
            (xmpp_id,),
        )
        return [r[0] for r in res.fetchall() if r[0] != xmpp_id]

    def attachment_get_legacy_id_for_xmpp_id(self, xmpp_id: str):
        res = self.cur.execute(
            "SELECT legacy_id FROM attachment_legacy_msg_id "
            "WHERE id = (SELECT legacy_msg_id FROM attachment_xmpp_ids WHERE xmpp_id = ?)",
            (xmpp_id,),
        )
        return first_of_tuple_or_none(res.fetchone())


class NickMixin(Base):
    def nick_get(self, jid: JID, user: "GatewayUser"):
        res = self.cur.execute(
            "SELECT nick FROM nick "
            "WHERE jid = ? "
            "AND user_id = (SELECT id FROM user WHERE jid = ?)",
            (str(jid), user.bare_jid),
        )
        return first_of_tuple_or_none(res.fetchone())

    def nick_store(self, jid: JID, nick: str, user: "GatewayUser"):
        self.cur.execute(
            "REPLACE INTO nick(jid, nick, user_id) "
            "VALUES (?,?,(SELECT id FROM user WHERE jid = ?))",
            (str(jid), nick, user.bare_jid),
        )
        self.con.commit()


class AvatarMixin(Base):
    def avatar_get(self, jid: JID):
        res = self.cur.execute(
            "SELECT cached_id FROM avatar WHERE jid = ?", (str(jid),)
        )
        return first_of_tuple_or_none(res.fetchone())

    def avatar_store(self, jid: JID, cached_id: Union[int, str]):
        self.cur.execute(
            "REPLACE INTO avatar(jid, cached_id) VALUES (?,?)", (str(jid), cached_id)
        )
        self.con.commit()

    def avatar_delete(self, jid: JID):
        self.cur.execute("DELETE FROM avatar WHERE jid = ?", (str(jid),))
        self.con.commit()


class PresenceMixin(Base):
    def __init__(self):
        super().__init__()
        self.__cur = cur = self.con.cursor()
        cur.row_factory = self.__row_factory  # type:ignore

    @staticmethod
    def __row_factory(
        _cur: sqlite3.Cursor,
        row: tuple[
            Optional[int],
            Optional[PresenceTypes],
            Optional[str],
            Optional[PresenceShows],
        ],
    ):
        if row[0] is not None:
            last_seen = datetime.fromtimestamp(row[0], tz=timezone.utc)
        else:
            last_seen = None
        return CachedPresence(last_seen, *row[1:])

    def presence_nuke(self):
        # useful for tests
        self.cur.execute("DELETE FROM presence")
        self.con.commit()

    def presence_store(self, jid: JID, presence: CachedPresence, user: "GatewayUser"):
        self.cur.execute(
            "REPLACE INTO presence(jid, last_seen, ptype, pstatus, pshow, user_id) "
            "VALUES (?,?,?,?,?,(SELECT id FROM user WHERE jid = ?))",
            (
                str(jid),
                presence[0].timestamp() if presence[0] else None,
                *presence[1:],
                user.bare_jid,
            ),
        )
        self.con.commit()

    def presence_delete(self, jid: JID, user: "GatewayUser"):
        self.cur.execute(
            "DELETE FROM presence WHERE (jid = ? and user_id = (SELECT id FROM user WHERE jid = ?))",
            (str(jid), user.bare_jid),
        )
        self.con.commit()

    def presence_get(self, jid: JID, user: "GatewayUser") -> Optional[CachedPresence]:
        return self.__cur.execute(
            "SELECT last_seen, ptype, pstatus, pshow FROM presence "
            "WHERE jid = ? AND user_id = (SELECT id FROM user WHERE jid = ?)",
            (str(jid), user.bare_jid),
        ).fetchone()


class UserMixin(Base):
    def user_store(self, user: "GatewayUser"):
        try:
            self.cur.execute("INSERT INTO user(jid) VALUES (?)", (user.bare_jid,))
        except sqlite3.IntegrityError:
            log.debug("User has already been added.")
        else:
            self.con.commit()

    def user_del(self, user: "GatewayUser"):
        self.cur.execute("DELETE FROM user WHERE jid = ?", (user.bare_jid,))
        self.con.commit()


def first_of_tuple_or_none(x: Optional[tuple]):
    if x is None:
        return None
    return x[0]


class SQLBiDict(Generic[KeyType, ValueType]):
    def __init__(
        self,
        table: str,
        key1: str,
        key2: str,
        user: "GatewayUser",
        sql: Optional[Base] = None,
        create_table=False,
        is_inverse=False,
    ):
        if sql is None:
            sql = db
        self.db = sql
        self.table = table
        self.key1 = key1
        self.key2 = key2
        self.user = user
        if create_table:
            sql.cur.execute(
                f"CREATE TABLE {table} (id "
                "INTEGER PRIMARY KEY,"
                "user_id INTEGER,"
                f"{key1} UNIQUE,"
                f"{key2} UNIQUE,"
                f"FOREIGN KEY(user_id) REFERENCES user(id))",
            )
        if is_inverse:
            return
        self.inverse = SQLBiDict[ValueType, KeyType](
            table, key2, key1, user, sql=sql, is_inverse=True
        )

    def __setitem__(self, key: KeyType, value: ValueType):
        self.db.cur.execute(
            f"REPLACE INTO {self.table}"
            f"(user_id, {self.key1}, {self.key2}) "
            "VALUES ((SELECT id FROM user WHERE jid = ?), ?, ?)",
            (self.user.bare_jid, key, value),
        )
        self.db.con.commit()

    def __getitem__(self, item: KeyType) -> ValueType:
        v = self.get(item)
        if v is None:
            raise KeyError(item)
        return v

    def __contains__(self, item: KeyType) -> bool:
        res = self.db.cur.execute(
            f"SELECT {self.key1} FROM {self.table} "
            f"WHERE {self.key1} = ? AND user_id = (SELECT id FROM user WHERE jid = ?)",
            (item, self.user.bare_jid),
        ).fetchone()
        return res is not None

    @lru_cache(100)
    def get(self, item: KeyType) -> Optional[ValueType]:
        res = self.db.cur.execute(
            f"SELECT {self.key2} FROM {self.table} "
            f"WHERE {self.key1} = ? AND user_id = (SELECT id FROM user WHERE jid = ?)",
            (item, self.user.bare_jid),
        ).fetchone()
        if res is None:
            return res
        return res[0]


class TemporaryDB(
    AvatarMixin, AttachmentMixin, NickMixin, MAMMixin, UserMixin, PresenceMixin
):
    pass


db = TemporaryDB()
log = logging.getLogger(__name__)



from slixmpp import JID

from ..util.db import log, user_store


class YesSet(set):
    """
    A pseudo-set which always test True for membership
    """

    def __contains__(self, item):
        log.debug("Test in")
        return True


class RosterBackend:
    """
    A pseudo-roster for the gateway component.

    If a user is in the user store, this will behave as if the user is part of the
    roster with subscription "both", and "none" otherwise.

    This is rudimentary but the only sane way I could come up with so far.
    """

    @staticmethod
    def entries(_owner_jid, _default=None):
        return YesSet()

    @staticmethod
    def save(_owner_jid, _jid, _item_state, _db_state):
        pass

    @staticmethod
    def load(_owner_jid, jid, _db_state):
        log.debug("Load %s", jid)
        user = user_store.get_by_jid(JID(jid))
        log.debug("User %s", user)
        if user is None:
            return {
                "name": "",
                "groups": [],
                "from": False,
                "to": False,
                "pending_in": False,
                "pending_out": False,
                "whitelisted": False,
                "subscription": "both",
            }
        else:
            return {
                "name": "",
                "groups": [],
                "from": True,
                "to": True,
                "pending_in": False,
                "pending_out": False,
                "whitelisted": False,
                "subscription": "none",
            }



from slixmpp.xmlstream import ElementBase


class Gateway(ElementBase):
    namespace = "jabber:iq:gateway"
    name = "query"
    plugin_attrib = "gateway"
    interfaces = {"desc", "prompt", "jid"}
    sub_interfaces = interfaces



from typing import TYPE_CHECKING

from slixmpp import CoroutineCallback, Iq, StanzaPath
from slixmpp.exceptions import XMPPError

if TYPE_CHECKING:
    from .base import BaseGateway


class MucAdmin:
    def __init__(self, xmpp: "BaseGateway"):
        self.xmpp = xmpp
        xmpp.register_handler(
            CoroutineCallback(
                "muc#admin",
                StanzaPath("iq@type=get/mucadmin_query"),
                self._handle_admin,  # type: ignore
            )
        )

    async def _handle_admin(self, iq: Iq):
        muc = await self.xmpp.get_muc_from_stanza(iq)

        affiliation = iq["mucadmin_query"]["item"]["affiliation"]

        if not affiliation:
            raise XMPPError("bad-request")

        reply = iq.reply()
        reply.enable("mucadmin_query")
        for participant in await muc.get_participants():
            if not participant.affiliation == affiliation:
                continue
            reply["mucadmin_query"].append(participant.mucadmin_item())
        reply.send()



"""
Slidge can be configured via CLI args, environment variables and/or INI files.

To use env vars, use this convention: ``--home-dir`` becomes ``HOME_DIR``.

Everything in ``/etc/slidge/conf.d/*`` is automatically used.
To use a plugin-specific INI file, put it in another dir,
and launch slidge with ``-c /path/to/plugin-specific.conf``.
Use the long version of the CLI arg without the double dash prefix inside this
INI file, eg ``debug=true``.

An example configuration file is available at
https://git.sr.ht/~nicoco/slidge/tree/master/item/dev/confs/slidge-example.ini
"""

import importlib
import logging
import os
import signal
from pathlib import Path

import configargparse

from slidge import BaseGateway
from slidge.core import config
from slidge.core.cache import avatar_cache
from slidge.migration import migrate
from slidge.util.conf import ConfigModule
from slidge.util.db import user_store
from slidge.util.util import get_version  # noqa: F401


class MainConfig(ConfigModule):
    def update_dynamic_defaults(self, args):
        # force=True is needed in case we call a logger before this is reached,
        # or basicConfig has no effect
        logging.basicConfig(level=args.loglevel, filename=args.log_file, force=True)

        if args.home_dir is None:
            args.home_dir = Path("/var/lib/slidge") / str(args.jid)

        if args.user_jid_validator is None:
            args.user_jid_validator = ".*@" + args.server


class SigTermInterrupt(Exception):
    pass


def get_configurator():
    p = configargparse.ArgumentParser(
        default_config_files=os.getenv(
            "SLIDGE_CONF_DIR", "/etc/slidge/conf.d/*.conf"
        ).split(":"),
        description=__doc__,
    )
    p.add_argument(
        "-c",
        "--config",
        help="Path to a INI config file.",
        env_var="SLIDGE_CONFIG",
        is_config_file=True,
    )
    p.add_argument(
        "-q",
        "--quiet",
        help="loglevel=WARNING",
        action="store_const",
        dest="loglevel",
        const=logging.WARNING,
        default=logging.INFO,
        env_var="SLIDGE_QUIET",
    )
    p.add_argument(
        "-d",
        "--debug",
        help="loglevel=DEBUG",
        action="store_const",
        dest="loglevel",
        const=logging.DEBUG,
        env_var="SLIDGE_DEBUG",
    )
    p.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s {__version__}",
    )
    configurator = MainConfig(config, p)
    return configurator


def get_parser():
    return get_configurator().parser


def configure():
    configurator = get_configurator()
    args, unknown_argv = configurator.set_conf()

    if not (h := config.HOME_DIR).exists():
        logging.info("Creating directory '%s'", h)
        h.mkdir()

    db_file = config.HOME_DIR / "slidge.db"
    user_store.set_file(db_file, args.secret_key)

    avatar_cache.set_dir(h / "slidge_avatars_v2")

    config.UPLOAD_REQUESTER = config.UPLOAD_REQUESTER or config.JID.bare

    return unknown_argv


def handle_sigterm(_signum, _frame):
    logging.info("Caught SIGTERM")
    raise SigTermInterrupt


def main():
    signal.signal(signal.SIGTERM, handle_sigterm)

    unknown_argv = configure()
    logging.info("Starting slidge version %s", __version__)

    legacy_module = importlib.import_module(config.LEGACY_MODULE)
    logging.debug("Legacy module: %s", dir(legacy_module))
    logging.info(
        "Starting legacy module: '%s' version %s",
        config.LEGACY_MODULE,
        getattr(legacy_module, "__version__", "No version"),
    )

    if plugin_config_obj := getattr(
        legacy_module, "config", getattr(legacy_module, "Config", None)
    ):
        logging.debug("Found a config object in plugin: %r", plugin_config_obj)
        ConfigModule.ENV_VAR_PREFIX += (
            f"_{config.LEGACY_MODULE.split('.')[-1].upper()}_"
        )
        logging.debug("Env var prefix: %s", ConfigModule.ENV_VAR_PREFIX)
        ConfigModule(plugin_config_obj).set_conf(unknown_argv)
    else:
        if unknown_argv:
            raise RuntimeError("Some arguments have not been recognized", unknown_argv)

    migrate()

    gateway: BaseGateway = BaseGateway.get_unique_subclass()()
    avatar_cache.http = gateway.http
    gateway.connect()

    return_code = 0
    try:
        gateway.loop.run_forever()
    except KeyboardInterrupt:
        logging.debug("Received SIGINT")
    except SigTermInterrupt:
        logging.debug("Received SIGTERM")
    except SystemExit as e:
        return_code = e.code  # type: ignore
        logging.debug("Exit called")
    except Exception as e:
        return_code = 2
        logging.exception("Exception in __main__")
        logging.exception(e)
    finally:
        if gateway.has_crashed:
            if return_code != 0:
                logging.warning("Return code has been set twice. Please report this.")
            return_code = 3
        if gateway.is_connected():
            logging.debug("Gateway is connected, cleaning up")
            gateway.shutdown()
            gateway.disconnect()
            gateway.loop.run_until_complete(gateway.disconnected)
        else:
            logging.debug("Gateway is not connected, no need to clean up")
        user_store.close()
        avatar_cache.close()
        gateway.loop.run_until_complete(gateway.http.close())
        logging.info("Successful clean shut down")
    logging.debug("Exiting with code %s", return_code)
    exit(return_code)


# this should be modified before publish, but if someone cloned from the repo,
# it can help
__version__ = get_version()


if __name__ == "__main__":
    main()



"""
Typing stuff
"""

from dataclasses import dataclass
from enum import IntEnum
from pathlib import Path
from typing import (
    IO,
    TYPE_CHECKING,
    Any,
    Generic,
    Hashable,
    Literal,
    NamedTuple,
    Optional,
    TypedDict,
    TypeVar,
    Union,
)

from slixmpp import Message, Presence
from slixmpp.types import PresenceShows

if TYPE_CHECKING:
    from ..contact import LegacyContact
    from ..core.pubsub import PepItem
    from ..core.session import BaseSession
    from ..group.participant import LegacyMUC, LegacyParticipant
    from .db import GatewayUser

    AnyBaseSession = BaseSession[Any, Any]
else:
    AnyBaseSession = None


class URL(str):
    pass


LegacyGroupIdType = TypeVar("LegacyGroupIdType", bound=Hashable)
"""
Type of the unique identifier for groups, usually a str or an int,
but anything hashable should work.
"""
LegacyMessageType = TypeVar("LegacyMessageType", bound=Hashable)
LegacyThreadType = TypeVar("LegacyThreadType", bound=Hashable)
LegacyUserIdType = TypeVar("LegacyUserIdType", bound=Hashable)

LegacyContactType = TypeVar("LegacyContactType", bound="LegacyContact")
LegacyMUCType = TypeVar("LegacyMUCType", bound="LegacyMUC")
LegacyParticipantType = TypeVar("LegacyParticipantType", bound="LegacyParticipant")

PepItemType = TypeVar("PepItemType", bound="PepItem")

Recipient = Union["LegacyMUC", "LegacyContact"]
RecipientType = TypeVar("RecipientType", bound=Recipient)
Sender = Union["LegacyContact", "LegacyParticipant"]
AvatarType = Union[bytes, str, Path]
LegacyFileIdType = Union[int, str]
AvatarIdType = Union[LegacyFileIdType, URL]

ChatState = Literal["active", "composing", "gone", "inactive", "paused"]
ProcessingHint = Literal["no-store", "markable", "store"]
Marker = Literal["acknowledged", "received", "displayed"]
FieldType = Literal[
    "boolean",
    "fixed",
    "text-single",
    "jid-single",
    "jid-multi",
    "list-single",
    "list-multi",
    "text-private",
]
MucAffiliation = Literal["owner", "admin", "member", "outcast", "none"]
MucRole = Literal["visitor", "participant", "moderator", "none"]


@dataclass
class MessageReference(Generic[LegacyMessageType]):
    """
    A "message reply", ie a "quoted message" (:xep:`0461`)

    At the very minimum, the legacy message ID attribute must be set, but to
    ensure that the quote is displayed in all XMPP clients, the author must also
    be set.
    The body is used as a fallback for XMPP clients that do not support :xep:`0461`
    of that failed to find the referenced message.
    """

    legacy_id: LegacyMessageType
    author: Optional[Union["GatewayUser", "LegacyParticipant", "LegacyContact"]] = None
    body: Optional[str] = None


@dataclass
class LegacyAttachment:
    """
    A file attachment to a message

    At the minimum, one of the ``path``, ``steam``, ``data`` or ``url`` attribute
    has to be set

    To be used with :meth:`.LegacyContact.send_files` or
    :meth:`.LegacyParticipant.send_files`
    """

    path: Optional[Union[Path, str]] = None
    name: Optional[Union[str]] = None
    stream: Optional[IO[bytes]] = None
    data: Optional[bytes] = None
    content_type: Optional[str] = None
    legacy_file_id: Optional[Union[str, int]] = None
    url: Optional[str] = None
    caption: Optional[str] = None
    """
    A caption for this specific image. For a global caption for a list of attachments,
    use the ``body`` parameter of :meth:`.AttachmentMixin.send_files`
    """

    def __post_init__(self):
        if not any(
            x is not None for x in (self.path, self.stream, self.data, self.url)
        ):
            raise TypeError("There is not data in this attachment", self)


class MucType(IntEnum):
    """
    The type of group, private, public, anonymous or not.
    """

    GROUP = 0
    """
    A private group, members-only and non-anonymous, eg a family group.
    """
    CHANNEL = 1
    """
    A public group, aka an anonymous channel.
    """
    CHANNEL_NON_ANONYMOUS = 2
    """
    A public group where participants' legacy IDs are visible to everybody.
    """


PseudoPresenceShow = Union[PresenceShows, Literal[""]]


class ResourceDict(TypedDict):
    show: PseudoPresenceShow
    status: str
    priority: int


MessageOrPresenceTypeVar = TypeVar(
    "MessageOrPresenceTypeVar", bound=Union[Message, Presence]
)


class LinkPreview(NamedTuple):
    about: str
    title: Optional[str]
    description: Optional[str]
    url: Optional[str]
    image: Optional[str]
    type: Optional[str]
    site_name: Optional[str]


class Mention(NamedTuple):
    contact: "LegacyContact"
    start: int
    end: int


class Hat(NamedTuple):
    uri: str
    title: str



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
from slixmpp.plugins import BasePlugin

from . import stanza


class LinkPreview(BasePlugin):
    name = "link_preview"
    description = "Sender-generated link previews"
    dependencies = set()
    stanza = stanza

    def plugin_init(self):
        stanza.register_plugin()



# Slixmpp: The Slick XMPP Library
# Copyright (C) 2012 Nathanael C. Fritz, Lance J.T. Stout
# This file is part of Slixmpp.
# See the file LICENSE for copying permission.
import logging
import ssl

from slixmpp.exceptions import XMPPError
from slixmpp.plugins import BasePlugin
from slixmpp.stanza import Iq, StreamFeatures
from slixmpp.xmlstream import JID, StanzaBase, register_stanza_plugin
from slixmpp.xmlstream.handler import CoroutineCallback
from slixmpp.xmlstream.matcher import StanzaPath

from . import stanza
from .stanza import Register, RegisterFeature

log = logging.getLogger(__name__)


# noinspection PyPep8Naming
class XEP_0077(BasePlugin):
    """
    XEP-0077: In-Band Registration

    Events:

    ::

        user_register           -- After successful validation and add to the user store
                                   in api["user_validate"]
        user_unregister         -- After successful user removal in api["user_remove"]
        user_modify             -- After successful user modify in api["user_modify"]

    Config:

    ::

        form_fields and form_instructions are only used if api["make_registration_form"] is
        not overridden; in this case form_fields MUST be None

    API:

    ::

        user_get(jid, node, ifrom, iq)
            Returns a dict-like object containing `form_fields` for this user or None
        user_remove(jid, node, ifrom, iq)
            Removes a user or raise KeyError in case the user is not found in the user store
        make_registration_form(self, jid, node, ifrom, iq)
            Returns an iq reply to a registration form request, pre-filled and with
            <registered/> in case the requesting entity is already registered to us
        user_validate((self, jid, node, ifrom, registration)
            Add the user to the user store or raise ValueError(msg) if any problem is encountered
            msg is sent back to the XMPP client as an error message.
        user_modify(jid, node, ifrom, iq)
            Modify the user in the user store or raise ValueError(msg) (similarly to user_validate)
    """

    name = "xep_0077"
    description = "XEP-0077: In-Band Registration (slidge)"
    dependencies = {"xep_0004", "xep_0066"}
    stanza = stanza
    default_config = {
        "create_account": True,
        "force_registration": False,
        "order": 50,
        "form_fields": {"username", "password"},
        "form_instructions": "Enter your credentials",
        "enable_subscription": True,
    }
    _user_store: dict[str, dict[str, str]]

    def plugin_init(self):
        register_stanza_plugin(StreamFeatures, RegisterFeature)
        register_stanza_plugin(Iq, Register)

        if self.xmpp.is_component:
            self.xmpp["xep_0030"].add_feature("jabber:iq:register")
            self.xmpp.register_handler(
                CoroutineCallback(
                    "registration",
                    StanzaPath(f"/iq@to={self.xmpp.boundjid.bare}/register"),
                    self._handle_registration,
                )
            )
            self._user_store = {}
            self.api.register(self._user_get, "user_get")
            self.api.register(self._user_remove, "user_remove")
            self.api.register(self._user_modify, "user_modify")
            self.api.register(self._make_registration_form, "make_registration_form")
            self.api.register(self._user_validate, "user_validate")
        else:
            self.xmpp.register_feature(
                "register",
                self._handle_register_feature,
                restart=False,
                order=self.order,
            )

        register_stanza_plugin(Register, self.xmpp["xep_0004"].stanza.Form)
        register_stanza_plugin(Register, self.xmpp["xep_0066"].stanza.OOB)

        self.xmpp.add_event_handler("connected", self._force_registration)

    def plugin_end(self):
        if not self.xmpp.is_component:
            self.xmpp.unregister_feature("register", self.order)

    def _user_get(self, _jid, _node, _ifrom, iq):
        return self._user_store.get(iq["from"].bare)

    def _user_remove(self, _jid, _node, _ifrom, iq):
        return self._user_store.pop(iq["from"].bare)

    async def _make_registration_form(self, _jid, _node, _ifrom, iq: Iq):
        reg = iq["register"]
        user = await self.api["user_get"](None, None, iq["from"], iq)

        if user is None:
            user = {}
        else:
            reg["registered"] = True

        reg["instructions"] = self.form_instructions

        for field in self.form_fields:
            data = user.get(field, "")
            if data:
                reg[field] = data
            else:
                # Add a blank field
                reg.add_field(field)

        reply = iq.reply()
        reply.set_payload(reg.xml)
        return reply

    def _user_validate(self, _jid, _node, ifrom, registration):
        self._user_store[ifrom.bare] = {
            key: registration[key] for key in self.form_fields
        }

    def _user_modify(self, _jid, _node, ifrom, registration):
        self._user_store[ifrom.bare] = {
            key: registration[key] for key in self.form_fields
        }

    async def _handle_registration(self, iq: StanzaBase):
        if iq["type"] == "get":
            if not self.enable_subscription:
                raise XMPPError(
                    "bad-request",
                    text="You must use adhoc commands to register to this gateway.",
                )
            await self._send_form(iq)
        elif iq["type"] == "set":
            form_dict = iq["register"]["form"].get_values() or iq["register"]

            if form_dict.get("remove"):
                try:
                    await self.api["user_remove"](None, None, iq["from"], iq)
                except KeyError:
                    _send_error(
                        iq,
                        "404",
                        "cancel",
                        "item-not-found",
                        "User not found",
                    )
                else:
                    reply = iq.reply()
                    reply.send()
                    self.xmpp.event("user_unregister", iq)
                return

            if not self.enable_subscription:
                raise XMPPError(
                    "bad-request",
                    text="You must use adhoc commands to register to this gateway.",
                )

            if self.form_fields is not None:
                for field in self.form_fields:
                    if not iq["register"][field]:
                        # Incomplete Registration
                        _send_error(
                            iq,
                            "406",
                            "modify",
                            "not-acceptable",
                            "Please fill in all fields.",
                        )
                        return

            user = await self.api["user_get"](None, None, iq["from"], iq)

            try:
                if user is None:
                    await self.api["user_validate"](None, None, iq["from"], form_dict)
                else:
                    await self.api["user_modify"](None, None, iq["from"], form_dict)
            except ValueError as e:
                _send_error(iq, "406", "modify", "not-acceptable", "\n".join(e.args))
                return

            reply = iq.reply()
            reply.send()

            if user is None:
                self.xmpp.event("user_register", iq)
            else:
                self.xmpp.event("user_modify", iq)

    async def _send_form(self, iq):
        reply = await self.api["make_registration_form"](None, None, iq["from"], iq)
        reply.send()

    def _force_registration(self, _event):
        if self.force_registration:
            self.xmpp.add_filter("in", self._force_stream_feature)

    def _force_stream_feature(self, stanza_):
        if isinstance(stanza_, StreamFeatures):
            if not self.xmpp.disable_starttls:
                if "starttls" not in self.xmpp.features:
                    return stanza_
                elif not isinstance(self.xmpp.socket, ssl.SSLSocket):
                    return stanza_
            if "mechanisms" not in self.xmpp.features:
                log.debug("Forced adding in-band registration stream feature")
                stanza_.enable("register")
                self.xmpp.del_filter("in", self._force_stream_feature)
        return stanza_

    async def _handle_register_feature(self, _features):
        if "mechanisms" in self.xmpp.features:
            # We have already logged in with an account
            return False

        if self.create_account and self.xmpp.event_handled("register"):
            form = await self.get_registration()
            await self.xmpp.event_async("register", form)
            return True
        return False

    def get_registration(self, jid=None, ifrom=None, timeout=None, callback=None):
        iq = self.xmpp.Iq()
        iq["type"] = "get"
        iq["to"] = jid
        iq["from"] = ifrom
        iq.enable("register")
        return iq.send(timeout=timeout, callback=callback)

    def cancel_registration(self, jid=None, ifrom=None, timeout=None, callback=None):
        iq = self.xmpp.Iq()
        iq["type"] = "set"
        iq["to"] = jid
        iq["from"] = ifrom
        iq["register"]["remove"] = True
        return iq.send(timeout=timeout, callback=callback)

    def change_password(
        self, password, jid=None, ifrom=None, timeout=None, callback=None
    ):
        iq = self.xmpp.Iq()
        iq["type"] = "set"
        iq["to"] = jid
        iq["from"] = ifrom
        if self.xmpp.is_component:
            ifrom = JID(ifrom)
            iq["register"]["username"] = ifrom.user
        else:
            iq["register"]["username"] = self.xmpp.boundjid.user
        iq["register"]["password"] = password
        return iq.send(timeout=timeout, callback=callback)


def _send_error(iq, code, error_type, name, text=""):
    # It would be nice to raise XMPPError but the iq payload
    # should include the register info
    reply = iq.reply()
    reply.set_payload(iq["register"].xml)
    reply.error()
    reply["error"]["code"] = code
    reply["error"]["type"] = error_type
    reply["error"]["condition"] = name
    reply["error"]["text"] = text
    reply.send()



from typing import TYPE_CHECKING

from slixmpp import CoroutineCallback, Iq, StanzaPath
from slixmpp.exceptions import XMPPError

from ...util.db import user_store

if TYPE_CHECKING:
    from .base import BaseGateway


class Mam:
    def __init__(self, xmpp: "BaseGateway"):
        self.xmpp = xmpp
        xmpp.register_handler(
            CoroutineCallback(
                "MAM_query",
                StanzaPath("iq@type=set/mam"),
                self.__handle_mam,  # type:ignore
            )
        )
        xmpp.register_handler(
            CoroutineCallback(
                "MAM_get_from",
                StanzaPath("iq@type=get/mam"),
                self.__handle_mam_get_form,  # type:ignore
            )
        )
        xmpp.register_handler(
            CoroutineCallback(
                "MAM_get_meta",
                StanzaPath("iq@type=get/mam_metadata"),
                self.__handle_mam_metadata,  # type:ignore
            )
        )

    async def __handle_mam(self, iq: Iq):
        muc = await self.xmpp.get_muc_from_stanza(iq)
        await muc.send_mam(iq)

    async def __handle_mam_get_form(self, iq: Iq):
        ito = iq.get_to()

        if ito == self.xmpp.boundjid.bare:
            raise XMPPError(
                text="No MAM on the component itself, use a JID with a resource"
            )

        ifrom = iq.get_from()
        user = user_store.get_by_jid(ifrom)
        if user is None:
            raise XMPPError("registration-required")

        session = self.xmpp.get_session_from_user(user)

        await session.bookmarks.by_jid(ito)

        reply = iq.reply()
        form = self.xmpp.plugin["xep_0004"].make_form()
        form.add_field(ftype="hidden", var="FORM_TYPE", value="urn:xmpp:mam:2")
        form.add_field(ftype="jid-single", var="with")
        form.add_field(ftype="text-single", var="start")
        form.add_field(ftype="text-single", var="end")
        form.add_field(ftype="text-single", var="before-id")
        form.add_field(ftype="text-single", var="after-id")
        form.add_field(ftype="boolean", var="include-groupchat")
        field = form.add_field(ftype="list-multi", var="ids")
        field["validate"]["datatype"] = "xs:string"
        field["validate"]["open"] = True
        reply["mam"].append(form)
        reply.send()

    async def __handle_mam_metadata(self, iq: Iq):
        muc = await self.xmpp.get_muc_from_stanza(iq)
        await muc.send_mam_metadata(iq)



from slixmpp.plugins import BasePlugin

from . import stanza


class XEP_0317(BasePlugin):
    """
    XEP-0317: Hats
    """

    name = "xep_0317"
    description = "XEP-0317: Hats"
    dependencies = {"xep_0045"}
    stanza = stanza

    def plugin_init(self):
        stanza.register()
